from itertools import product
from typing import ClassVar
from unittest.mock import MagicMock, patch

import pytest
from _pytest.mark import ParameterSet

from cognite_toolkit._cdf_tk.utils.aggregators import (
    AssetAggregator,
    EventAggregator,
    FileAggregator,
    MetadataAggregator,
    SequenceAggregator,
    TimeSeriesAggregator,
)
from cognite_toolkit._cdf_tk.utils.cdf import label_count, metadata_key_counts
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.approval_client.client import LookUpAPIMock


class TestAggregators:
    @pytest.mark.parametrize(
        "aggregator_class, hierarchy, data_set_external_id, expected_filter",
        [
            (filter_cls, *args)
            for filter_cls, args in product(
                [AssetAggregator, FileAggregator, TimeSeriesAggregator, SequenceAggregator, EventAggregator],
                [
                    (None, None, None),
                    ("hierarchy", None, {"assetSubtreeIds": [{"externalId": "hierarchy"}]}),
                    ("", None, {"assetSubtreeIds": [{"externalId": ""}]}),
                    (None, "data_set", {"dataSetIds": [{"externalId": "data_set"}]}),
                    (None, "", {"dataSetIds": [{"externalId": ""}]}),
                    (
                        "hierarchy",
                        "data_set",
                        {"assetSubtreeIds": [{"externalId": "hierarchy"}], "dataSetIds": [{"externalId": "data_set"}]},
                    ),
                    (
                        "hierarchy",
                        "",
                        {"assetSubtreeIds": [{"externalId": "hierarchy"}], "dataSetIds": [{"externalId": ""}]},
                    ),
                    (
                        "",
                        "data_set",
                        {"assetSubtreeIds": [{"externalId": ""}], "dataSetIds": [{"externalId": "data_set"}]},
                    ),
                    ("", "", {"assetSubtreeIds": [{"externalId": ""}], "dataSetIds": [{"externalId": ""}]}),
                    ([], [], None),
                    (["hierarchy"], [], {"assetSubtreeIds": [{"externalId": "hierarchy"}]}),
                    ([], ["data_set"], {"dataSetIds": [{"externalId": "data_set"}]}),
                    (
                        ["hierarchy"],
                        ["data_set"],
                        {"assetSubtreeIds": [{"externalId": "hierarchy"}], "dataSetIds": [{"externalId": "data_set"}]},
                    ),
                    (
                        ["hierarchy"],
                        [""],
                        {"assetSubtreeIds": [{"externalId": "hierarchy"}], "dataSetIds": [{"externalId": ""}]},
                    ),
                    ([], ["data_set", ""], {"dataSetIds": [{"externalId": "data_set"}, {"externalId": ""}]}),
                ],
            )
        ],
    )
    def test_create_filter_method(
        self,
        aggregator_class: type[MetadataAggregator],
        hierarchy: str | None,
        data_set_external_id: str | None,
        expected_filter: dict[str, object] | None,
    ) -> None:
        actual_filter = aggregator_class.create_filter(hierarchy, data_set_external_id)
        assert (actual_filter is None and expected_filter is None) or actual_filter.dump() == expected_filter, (
            f"Failed with {hierarchy}, {data_set_external_id} -> {actual_filter.dump()} != {expected_filter}"
        )

    hierarchy_dataset_combinations: ClassVar[list[ParameterSet]] = [
        pytest.param(None, None, None, None, id="No hierarchy and no data sets"),
        pytest.param("hierarchy", None, [LookUpAPIMock.create_id("hierarchy")], None, id="Hierarchy only"),
        pytest.param(None, "data_set", None, [LookUpAPIMock.create_id("data_set")], id="Data set only"),
        pytest.param(
            "hierarchy",
            "data_set",
            [LookUpAPIMock.create_id("hierarchy")],
            [LookUpAPIMock.create_id("data_set")],
            id="Hierarchy and data set",
        ),
        pytest.param(
            ["hierarchy1", "hierarchy2"],
            ["data_set1", "data_set2"],
            sorted([LookUpAPIMock.create_id("hierarchy1"), LookUpAPIMock.create_id("hierarchy2")]),
            sorted([LookUpAPIMock.create_id("data_set1"), LookUpAPIMock.create_id("data_set2")]),
            id="Multiple hierarchies and data sets",
        ),
    ]

    @pytest.mark.parametrize("hierarchy, data_sets, hierarchy_ids, data_set_ids", hierarchy_dataset_combinations)
    def test_metadata_keys_count(
        self,
        hierarchy: str | list[str] | None,
        data_sets: str | list[str] | None,
        hierarchy_ids: list[int] | None,
        data_set_ids: list[int] | None,
        toolkit_client_approval: ApprovalToolkitClient,
    ) -> None:
        aggregator = AssetAggregator(toolkit_client_approval.mock_client)

        metadata_key_counts_mock = MagicMock(spec=metadata_key_counts)
        metadata_key_counts_mock.return_value = [("key1", 10), ("key2", 5), ("key3", 15)]

        with patch(f"{AssetAggregator.__module__}.metadata_key_counts", metadata_key_counts_mock):
            result = aggregator.metadata_key_count(hierarchy, data_sets)

        assert result == 3
        metadata_key_counts_mock.assert_called_once_with(
            toolkit_client_approval.mock_client, "assets", hierarchies=hierarchy_ids, data_sets=data_set_ids
        )

    @pytest.mark.parametrize("hierarchy, data_sets, hierarchy_ids, data_set_ids", hierarchy_dataset_combinations)
    def test_label_count(
        self,
        hierarchy: str | list[str] | None,
        data_sets: str | list[str] | None,
        hierarchy_ids: list[int] | None,
        data_set_ids: list[int] | None,
        toolkit_client_approval: ApprovalToolkitClient,
    ) -> None:
        aggregator = AssetAggregator(toolkit_client_approval.mock_client)

        label_count_mock = MagicMock(spec=label_count)
        label_count_mock.return_value = [("label1", 42), ("label2", 10)]

        with patch(f"{AssetAggregator.__module__}.label_count", label_count_mock):
            result = aggregator.label_count(hierarchy, data_sets)

        assert result == 2
        label_count_mock.assert_called_once_with(
            toolkit_client_approval.mock_client, "assets", hierarchies=hierarchy_ids, data_sets=data_set_ids
        )



from pydantic import Field, JsonValue

from .base import ToolkitResource


class ExtractionPipelineConfigYAML(ToolkitResource):
    external_id: str = Field(
        description="External ID of the extraction pipeline this configuration revision belongs to.",
        min_length=1,
        max_length=255,
    )
    description: str | None = Field(None, description="A description of this configuration revision.")
    config: str | dict[str, JsonValue] = Field(
        description="Configuration content. Can be either a JSON string or a dictionary."
    )



import functools
import platform

from cognite_toolkit import _version


def get_current_toolkit_version() -> str:
    return _version.__version__


@functools.lru_cache(maxsize=1)
def get_user_agent() -> str:
    toolkit_version = f"CogniteToolkit/{get_current_toolkit_version()}"
    python_version = (
        f"{platform.python_implementation()}/{platform.python_version()} "
        f"({platform.python_build()};{platform.python_compiler()})"
    )
    os_version_info = [platform.release(), platform.machine(), platform.architecture()[0]]
    os_version_info = [s for s in os_version_info if s]  # Ignore empty strings
    os_version_info_str = "-".join(os_version_info)
    operating_system = f"{platform.system()}/{os_version_info_str}"

    return f"{toolkit_version} {python_version} {operating_system}"



from collections.abc import Iterable
from pathlib import Path
from typing import Any

import pytest

from cognite_toolkit._cdf_tk.resource_classes.infield_v1 import InfieldV1YAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.data import COMPLETE_ORG_ALPHA_FLAGS
from tests.test_unit.utils import find_resources


def invalid_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "InfieldApp"},
        {"Missing required field: 'externalId'"},
        id="Missing required field: externalId",
    )
    yield pytest.param(
        {
            "externalId": 123,  # Should be string
            "name": 456,  # Should be string
            "appDataSpaceVersion": [],  # Should be string
            "dataSetId": "invalid",  # Unknown field
        },
        {
            "In field externalId input should be a valid string. Got 123 of type int. Hint: Use double quotes to force string.",
            "In field name input should be a valid string. Got 456 of type int. Hint: Use double quotes to force string.",
            "In field appDataSpaceVersion input should be a valid string. Got [] of type list. Hint: Use double quotes to force string.",
            "Unused field: 'dataSetId'",
        },
        id="Multiple type validation errors",
    )
    yield pytest.param(
        {
            "externalId": "test-config",
            "featureConfiguration": {
                "rootLocationConfigurations": [
                    {
                        "dataSetExternalId": True,  # Should be string
                        "templateAdmins": "not-a-list",  # Should be list
                        "threeDConfiguration": {
                            "fullWeightModels": [
                                {
                                    "revisionId": "not-an-int",  # Should be int
                                    "modelId": "also-not-an-int",  # Should be int
                                }
                            ]
                        },
                        "featureToggles": {
                            "threeD": "not-a-bool",  # Should be bool
                            "observations": {
                                "isEnabled": "not-a-bool",  # Should be bool
                            },
                        },
                    }
                ]
            },
        },
        {
            "In featureConfiguration.rootLocationConfigurations[1].dataSetExternalId "
            "input should be a valid string. Got True of type bool. Hint: Use double "
            "quotes to force string.",
            "In "
            "featureConfiguration.rootLocationConfigurations[1].featureToggles.observations.isEnabled "
            "input should be a valid boolean. Got 'not-a-bool' of type str.",
            "In featureConfiguration.rootLocationConfigurations[1].featureToggles.threeD "
            "input should be a valid boolean. Got 'not-a-bool' of type str.",
            "In featureConfiguration.rootLocationConfigurations[1].templateAdmins input "
            "should be a valid list. Got 'not-a-list'.",
            "In "
            "featureConfiguration.rootLocationConfigurations[1].threeDConfiguration.fullWeightModels[1].modelId "
            "input should be a valid integer. Got 'also-not-an-int' of type str.",
            "In "
            "featureConfiguration.rootLocationConfigurations[1].threeDConfiguration.fullWeightModels[1].revisionId "
            "input should be a valid integer. Got 'not-an-int' of type str.",
        },
        id="Nested structure validation errors",
    )
    yield pytest.param(
        {
            "externalId": "test-config",
            "featureConfiguration": {
                "rootLocationConfigurations": [
                    {
                        "observations": {
                            "type": {
                                "options": [
                                    {
                                        "id": 123,  # Should be string
                                        "value": [],  # Should be string
                                        "label": {"invalid": "object"},  # Should be string
                                    }
                                ]
                            }
                        },
                        "dataFilters": {
                            "general": {
                                "dataSetExternalIds": [1, 32],  # Wrong field name, should be data_set_ids
                                "spaces": 42,  # Should be list of strings
                            }
                        },
                    }
                ]
            },
        },
        {
            "In "
            "featureConfiguration.rootLocationConfigurations[1].dataFilters.general.dataSetExternalIds[1] "
            "input should be a valid string. Got 1 of type int. Hint: Use double quotes "
            "to force string.",
            "In "
            "featureConfiguration.rootLocationConfigurations[1].dataFilters.general.dataSetExternalIds[2] "
            "input should be a valid string. Got 32 of type int. Hint: Use double quotes "
            "to force string.",
            "In "
            "featureConfiguration.rootLocationConfigurations[1].dataFilters.general.spaces "
            "input should be a valid list. Got 42.",
            "In "
            "featureConfiguration.rootLocationConfigurations[1].observations.type.options[1].id "
            "input should be a valid string. Got 123 of type int. Hint: Use double quotes "
            "to force string.",
            "In "
            "featureConfiguration.rootLocationConfigurations[1].observations.type.options[1].label "
            "input should be a valid string. Got {'invalid': 'object'} of type dict. "
            "Hint: Use double quotes to force string.",
            "In "
            "featureConfiguration.rootLocationConfigurations[1].observations.type.options[1].value "
            "input should be a valid string. Got [] of type list. Hint: Use double quotes "
            "to force string.",
        },
        id="Deep nested validation with complex structures",
    )


class TestInfieldV1YAML:
    @pytest.mark.parametrize(
        "data", list(find_resources("InfieldV1", resource_dir="cdf_applications", base=COMPLETE_ORG_ALPHA_FLAGS))
    )
    def test_load_valid(self, data: dict[str, Any]) -> None:
        loaded = InfieldV1YAML.model_validate(data)

        dumped = loaded.model_dump(exclude_unset=True, by_alias=True)
        assert dumped == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_test_cases()))
    def test_invalid_error_messages(self, data: dict[str, Any], expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, InfieldV1YAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from typing import Any, Literal

from pydantic import Field

from .base import ToolkitResource


class AssetYAML(ToolkitResource):
    external_id: str = Field(description="External ID provided by the client.", max_length=255)
    name: str | None = Field(None, description="The name of the asset.", min_length=1, max_length=140)
    parent_external_id: str | None = Field(
        None, description="The external ID of the parent of the node.", max_length=255
    )
    description: str | None = Field(None, description="The description of the asset.", max_length=500)
    metadata: dict[str, str] | None = Field(None, description="Custom, application-specific metadata.", max_length=256)
    data_set_external_id: str | None = Field(None, description="The external id of the dataset this asset belongs to.")
    source: str | None = Field(None, description="The source of the asset.", max_length=128)
    labels: list[str | dict[Literal["externalId"], str]] | None = Field(
        None, description="A list of labels associated with this resource.", max_length=10
    )
    geo_location: dict[str, Any] | None = Field(None, description="The geographic metadata of the asset")



from abc import ABC, abstractmethod
from typing import Generic

from cognite.client.data_classes._base import (
    T_CogniteResource,
)
from cognite.client.data_classes.data_modeling import DirectRelationReference, InstanceApply, View, ViewId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.migration import ResourceViewMapping
from cognite_toolkit._cdf_tk.commands._migrate.conversion import asset_centric_to_dm
from cognite_toolkit._cdf_tk.commands._migrate.data_classes import AssetCentricMapping
from cognite_toolkit._cdf_tk.commands._migrate.issues import ConversionIssue, MigrationIssue
from cognite_toolkit._cdf_tk.commands._migrate.selectors import AssetCentricMigrationSelector
from cognite_toolkit._cdf_tk.constants import MISSING_INSTANCE_SPACE
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.storageio._base import T_Selector, T_WriteCogniteResource
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.useful_types import T_AssetCentricResource


class DataMapper(Generic[T_Selector, T_CogniteResource, T_WriteCogniteResource], ABC):
    def prepare(self, source_selector: T_Selector) -> None:
        """Prepare the data mapper with the given source selector.

        Args:
            source_selector: The selector for the source data.

        """
        # Override in subclass if needed.
        pass

    @abstractmethod
    def map(self, source: T_CogniteResource) -> tuple[T_WriteCogniteResource | None, MigrationIssue]:
        """Map a chunk of source data to the target format.

        Args:
            source: The source data chunk to be mapped.

        Returns:
            A tuple containing the mapped data and a list of any issues encountered during mapping.

        """
        raise NotImplementedError("Subclasses must implement this method.")


class AssetCentricMapper(
    DataMapper[AssetCentricMigrationSelector, AssetCentricMapping[T_AssetCentricResource], InstanceApply]
):
    def __init__(self, client: ToolkitClient) -> None:
        self.client = client
        self._ingestion_view_by_id: dict[ViewId, View] = {}
        self._view_mapping_by_id: dict[str, ResourceViewMapping] = {}
        # This is used to keep track of already mapped assets, such that we can creat direct relations
        # to them from files, events, and time series.
        self._asset_mapping_by_id: dict[int, DirectRelationReference] = {}
        self._source_system_mapping_by_id: dict[str, DirectRelationReference] = {}

    def prepare(self, source_selector: AssetCentricMigrationSelector) -> None:
        ingestion_view_ids = source_selector.get_ingestion_mappings()
        ingestion_views = self.client.migration.resource_view_mapping.retrieve(ingestion_view_ids)
        self._view_mapping_by_id = {view.external_id: view for view in ingestion_views}
        missing_mappings = set(ingestion_view_ids) - set(self._view_mapping_by_id.keys())
        if missing_mappings:
            raise ToolkitValueError(
                f"The following ingestion views were not found: {humanize_collection(missing_mappings)}"
            )

        view_ids = list({view.view_id for view in ingestion_views})
        views = self.client.data_modeling.views.retrieve(view_ids)
        self._ingestion_view_by_id = {view.as_id(): view for view in views}
        missing_views = set(view_ids) - set(self._ingestion_view_by_id.keys())
        if missing_views:
            raise ToolkitValueError(
                f"The following ingestion views were not found in Data Modeling: {humanize_collection(missing_views)}"
            )
        # We just look-up all source system for now. This can be optimized to only
        # look-up the ones that are actually used in the ingestion views. However, SourceSystem is typically in
        # the order ~10 instances, so this is not a big deal for now. See task CDF-25974.
        source_systems = self.client.migration.created_source_system.list(limit=-1)
        self._source_system_mapping_by_id = {
            source_system.source: source_system.as_direct_relation_reference() for source_system in source_systems
        }

        # We look-up all existing asset mappings to be able to create direct relations to already mapped assets.
        # This is needed in case the migration is run multiple times, or if assets are mapped
        asset_mappings = self.client.migration.instance_source.list(resource_type="asset", limit=-1)
        self._asset_mapping_by_id = {mapping.id_: mapping.as_direct_relation_reference() for mapping in asset_mappings}

    def map(self, source: AssetCentricMapping[T_AssetCentricResource]) -> tuple[InstanceApply | None, ConversionIssue]:
        """Map a chunk of asset-centric data to InstanceApplyList format."""
        mapping = source.mapping
        ingestion_view = mapping.get_ingestion_view()
        try:
            view_source = self._view_mapping_by_id[ingestion_view]
            view_properties = self._ingestion_view_by_id[view_source.view_id].properties
        except KeyError as e:
            raise RuntimeError(
                f"Failed to lookup mapping or view for ingestion view '{ingestion_view}'. Did you forget to call .prepare()?"
            ) from e
        instance, conversion_issue = asset_centric_to_dm(
            source.resource,
            instance_id=mapping.instance_id,
            view_source=view_source,
            view_properties=view_properties,
            asset_instance_id_by_id=self._asset_mapping_by_id,
            source_instance_id_by_external_id=self._source_system_mapping_by_id,
            file_instance_id_by_id={},  # Todo implement file direct relations
        )
        if mapping.instance_id.space == MISSING_INSTANCE_SPACE:
            conversion_issue.missing_instance_space = f"Missing instance space for dataset ID {mapping.data_set_id!r}"

        if mapping.resource_type == "asset":
            self._asset_mapping_by_id[mapping.id] = DirectRelationReference(
                space=mapping.instance_id.space, external_id=mapping.instance_id.external_id
            )
        return instance, conversion_issue



from cognite.client.data_classes.data_modeling import ViewId

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError


def parse_view_str(view_str: str) -> ViewId:
    """Parse a view string into a ViewId.

    Args:
        view_str (str): The view string to parse.

    Returns:
        ViewId: The parsed ViewId.

    Raises:
        ToolkitValueError: If the view string is not in a valid format.

    >>> parse_view_str("my_space:my_view/v1")
    ViewId(space='my_space', external_id='my_view', version='v1')
    """
    space_and_rest = view_str.split(":", 1)
    if len(space_and_rest) != 2 or not space_and_rest[0] or not space_and_rest[1]:
        raise ToolkitValueError(
            f"Invalid view string format: '{view_str}'. Expected format 'space:externalId/version'."
        )
    space = space_and_rest[0]
    external_id_and_version = space_and_rest[1].rsplit("/", 1)
    if len(external_id_and_version) != 2 or not external_id_and_version[0] or not external_id_and_version[1]:
        raise ToolkitValueError(
            f"Invalid view string format: '{view_str}'. Expected format 'space:externalId/version'."
        )
    external_id = external_id_and_version[0]
    version = external_id_and_version[1]
    return ViewId(space=space, external_id=external_id, version=version)



from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence, Sized
from dataclasses import dataclass
from typing import ClassVar, Generic, TypeVar

from cognite.client.data_classes._base import T_CogniteResource

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.fileio import FileReader, SchemaColumn
from cognite_toolkit._cdf_tk.utils.fileio._readers import TableReader
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, HTTPMessage, ItemsRequest
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal, T_WriteCogniteResource

from .selectors import DataSelector


@dataclass
class StorageIOConfig:
    kind: str
    folder_name: str
    value: JsonVal
    filename: str | None = None


T_Selector = TypeVar("T_Selector", bound=DataSelector)


@dataclass
class Page(Generic[T_CogniteResource], Sized):
    worker_id: str
    items: Sequence[T_CogniteResource]
    next_cursor: str | None = None

    def __len__(self) -> int:
        return len(self.items)


@dataclass
class UploadItem(Generic[T_WriteCogniteResource]):
    """An item to be uploaded to CDF, consisting of a source ID and the writable Cognite resource.

    Attributes:
        source_id: The source identifier for the item. For example, the line number in a CSV file.
        item: The writable Cognite resource to be uploaded.
    """

    source_id: str
    item: T_WriteCogniteResource

    def as_id(self) -> str:
        return self.source_id

    def dump(self) -> JsonVal:
        return self.item.dump(camel_case=True)


class StorageIO(ABC, Generic[T_Selector, T_CogniteResource]):
    """This is a base class for all storage classes in Cognite Toolkit

    It defines the interface for downloading data from CDF. Note this can also be used for multiple
    types of resources, for example, a hierarchy of assets/files/events/time series.

    Attributes
        SUPPORTED_DOWNLOAD_FORMATS: A set of formats that the storage item supports for downloading.
        SUPPORTED_COMPRESSIONS: A set of compression formats that the storage item supports.
        CHUNK_SIZE: The size of the data chunks to be processed during download and upload operations.

    Args:
        client: An instance of ToolkitClient to interact with the CDF API.
    """

    SUPPORTED_DOWNLOAD_FORMATS: ClassVar[frozenset[str]]
    SUPPORTED_COMPRESSIONS: ClassVar[frozenset[str]]
    CHUNK_SIZE: ClassVar[int]
    BASE_SELECTOR: ClassVar[type[DataSelector]]

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    @abstractmethod
    def as_id(self, item: T_CogniteResource) -> str:
        """Convert an item to its corresponding ID.
        Args:
            item: The item to convert.
        Returns:
            The ID corresponding to the item.
        """
        raise NotImplementedError()

    @abstractmethod
    def stream_data(self, selector: T_Selector, limit: int | None = None) -> Iterable[Page]:
        """Download items from the storage given the selection criteria.

        Args:
            selector: The selection criteria to filter the items to download.
            limit: Optional limit on the number of items to download.

        Returns:
            An iterable of writable Cognite resource lists.
        """
        raise NotImplementedError()

    @abstractmethod
    def count(self, selector: T_Selector) -> int | None:
        """Count the number of items in the storage that match the given selector.

        Args:
            selector: The selection criteria to filter the items to count.

        Returns:
            The number of items that match the selection criteria, or None if counting is not supported.
        """
        raise NotImplementedError()

    @abstractmethod
    def data_to_json_chunk(
        self, data_chunk: Sequence[T_CogniteResource], selector: T_Selector | None = None
    ) -> list[dict[str, JsonVal]]:
        """Convert a chunk of data to a JSON-compatible format.

        Args:
            data_chunk: The chunk of data to convert, which should be a writable Cognite resource list.
            selector: Optional selection criteria to identify the data. This is required for some storage types.

        Returns:
            A list of dictionaries representing the data in a JSON-compatible format.

        """
        raise NotImplementedError()


class UploadableStorageIO(
    Generic[T_Selector, T_CogniteResource, T_WriteCogniteResource], StorageIO[T_Selector, T_CogniteResource], ABC
):
    """A base class for storage items that support uploading data to CDF.

    Attributes:
        KIND: The kind of storage item (e.g., "RAW", "AssetCentric").
        SUPPORTED_READ_FORMATS: A set of formats that the storage item supports for reading.
        UPLOAD_ENDPOINT: The API endpoint for uploading data to the storage item.
        UPLOAD_EXTRA_ARGS: Additional arguments to include in the upload request.
    """

    KIND: ClassVar[str]
    SUPPORTED_READ_FORMATS: ClassVar[frozenset[str]]
    UPLOAD_ENDPOINT: ClassVar[str]
    UPLOAD_EXTRA_ARGS: ClassVar[Mapping[str, JsonVal] | None] = None

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[T_WriteCogniteResource]],
        http_client: HTTPClient,
        selector: T_Selector | None = None,
    ) -> Sequence[HTTPMessage]:
        """Upload a chunk of data to the storage using a custom HTTP client.
        This ensures that even if one item in the chunk fails, the rest will still be uploaded.

        This assumes that the data_chunk is respecting the CHUNK_SIZE of the storage.

        Args:
            data_chunk: The chunk of data to upload, which should be a list of writable Cognite resources.
            http_client: The custom HTTP client to use for the upload.
            selector: Optional selection criteria to identify where to upload the data.
        """
        if not hasattr(self, "UPLOAD_ENDPOINT"):
            raise ToolkitNotImplementedError(f"Upload not implemented for {self.KIND} storage.")
        if len(data_chunk) > self.CHUNK_SIZE:
            raise ValueError(f"Data chunk size {len(data_chunk)} exceeds the maximum CHUNK_SIZE of {self.CHUNK_SIZE}.")

        config = http_client.config
        return http_client.request_with_retries(
            message=ItemsRequest(
                endpoint_url=config.create_api_url(self.UPLOAD_ENDPOINT),
                method="POST",
                items=list(data_chunk),
                extra_body_fields=dict(self.UPLOAD_EXTRA_ARGS or {}),
            )
        )

    def json_chunk_to_data(
        self, data_chunk: list[tuple[str, dict[str, JsonVal]]]
    ) -> Sequence[UploadItem[T_WriteCogniteResource]]:
        """Convert a JSON-compatible chunk of data back to a writable Cognite resource list.

        Args:
            data_chunk: A list of tuples, each containing a source ID and a dictionary representing
                the data in a JSON-compatible format.
        Returns:
            A writable Cognite resource list representing the data.
        """
        result: list[UploadItem[T_WriteCogniteResource]] = []
        for source_id, item_json in data_chunk:
            item = self.json_to_resource(item_json)
            result.append(UploadItem(source_id=source_id, item=item))
        return result

    @abstractmethod
    def json_to_resource(self, item_json: dict[str, JsonVal]) -> T_WriteCogniteResource:
        """Convert a JSON-compatible dictionary back to a writable Cognite resource.

        Args:
            item_json: A dictionary representing the data in a JSON-compatible format.
        Returns:
            A writable Cognite resource representing the data.
        """
        raise NotImplementedError()

    @classmethod
    def read_chunks(cls, reader: FileReader) -> Iterable[list[tuple[str, dict[str, JsonVal]]]]:
        data_name = "row" if isinstance(reader, TableReader) else "line"
        # Include name of line for better error messages
        iterable = ((f"{data_name} {line_no}", item) for line_no, item in reader.read_chunks_with_line_numbers())

        yield from chunker(iterable, cls.CHUNK_SIZE)


class TableUploadableStorageIO(UploadableStorageIO[T_Selector, T_CogniteResource, T_WriteCogniteResource], ABC):
    """A base class for storage items that support uploading data with table schemas."""

    def rows_to_data(
        self, rows: list[tuple[str, dict[str, JsonVal]]], selector: T_Selector | None = None
    ) -> Sequence[UploadItem[T_WriteCogniteResource]]:
        """Convert a row-based JSON-compatible chunk of data back to a writable Cognite resource list.

        Args:
            rows: A list of tuples, each containing a source ID and a dictionary representing
                the data in a JSON-compatible format.
            selector: Optional selection criteria to identify where to upload the data. This is required for some storage types.

        Returns:
            A writable Cognite resource list representing the data.
        """
        result: list[UploadItem[T_WriteCogniteResource]] = []
        for source_id, row in rows:
            item = self.row_to_resource(source_id, row, selector=selector)
            result.append(UploadItem(source_id=source_id, item=item))
        return result

    @abstractmethod
    def row_to_resource(
        self, source_id: str, row: dict[str, JsonVal], selector: T_Selector | None = None
    ) -> T_WriteCogniteResource:
        """Convert a row-based JSON-compatible dictionary back to a writable Cognite resource.

        Args:
            row: A dictionary representing the data in a JSON-compatible format.
            source_id: The source identifier for the item. For example, the line number in a CSV file.
            selector: Optional selection criteria to identify where to upload the data. This is required for some storage types.
        Returns:
            A writable Cognite resource representing the data.
        """
        raise NotImplementedError()


class ConfigurableStorageIO(StorageIO[T_Selector, T_CogniteResource], ABC):
    """A base class for storage items that support configurations for different storage items."""

    @abstractmethod
    def configurations(self, selector: T_Selector) -> Iterable[StorageIOConfig]:
        """Return configurations for the storage item."""
        raise NotImplementedError()


class TableStorageIO(StorageIO[T_Selector, T_CogniteResource], ABC):
    """A base class for storage items that support table schemas."""

    @abstractmethod
    def get_schema(self, selector: T_Selector) -> list[SchemaColumn]:
        """Get the schema of the table associated with the given selector.

        Args:
            selector: The selection criteria to identify the data.

        Returns:
            A list of SchemaColumn objects representing the schema of the table.

        """
        raise NotImplementedError()

    @abstractmethod
    def data_to_row(
        self, data_chunk: Sequence[T_CogniteResource], selector: T_Selector | None = None
    ) -> list[dict[str, JsonVal]]:
        """Convert a chunk of data to a row-based JSON-compatible format.

        Args:
            data_chunk: The chunk of data to convert, which should be a writable Cognite resource list.
            selector: Optional selection criteria to identify the data. This is required for some storage types.

        Returns:
            A list of dictionaries representing the data in a JSON-compatible format.

        """
        raise NotImplementedError()



from collections.abc import Sequence
from typing import Literal, TypeAlias, overload

from cognite.client.data_classes.capabilities import (
    AllScope,
    AssetsAcl,
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    DataSetScope,
    EventsAcl,
    ExtractionPipelinesAcl,
    FilesAcl,
    LabelsAcl,
    RelationshipsAcl,
    SequencesAcl,
    ThreeDAcl,
    TimeSeriesAcl,
    TransformationsAcl,
    WorkflowOrchestrationAcl,
)

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.exceptions import AuthorizationError
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection

Action: TypeAlias = Literal["read", "write"]


class ValidateAccess:
    def __init__(self, client: ToolkitClient, default_operation: str) -> None:
        self.client = client
        self.default_operation = default_operation

    def data_model(
        self, action: Sequence[Action], spaces: set[str] | None = None, operation: str | None = None
    ) -> list[str] | None:
        """Validate access to data models.

        Args:
            action (Sequence[Action]): The actions to validate access for.
            spaces (Set[str] | None): The space IDs to check access for. If None, checks access for all spaces.
            operation (str | None): The operation being performed, used for error messages.

        Returns:
            list[str] | None: Returns a list of space IDs if access is limited to these spaces, or None if access is granted to all spaces.

        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected number of data model scopes.
            AuthorizationError: If the user does not have permission to perform the specified action on the given space.
        """
        operation = operation or self.default_operation
        model_scopes = self._get_scopes(
            action, DataModelsAcl.Action.Read, DataModelsAcl.Action.Write, operation, "data models"
        )
        if len(model_scopes) != 1:
            raise ValueError(f"Unexpected number of data model scopes: {len(model_scopes)}. Expected 1 scope.")
        model_scope = model_scopes[0]
        if isinstance(model_scope, DataModelsAcl.Scope.All):
            return None
        elif isinstance(model_scope, DataModelsAcl.Scope.SpaceID):
            if spaces is None:
                return model_scope.space_ids
            if missing := spaces - set(model_scope.space_ids):
                raise AuthorizationError(
                    f"You have no permission to {humanize_collection(action)} the {humanize_collection(missing)!r} "
                    f"space(s). This is required to {operation}."
                )
            return None
        else:
            raise ValueError(f"Unexpected data model scope type: {type(model_scope)}. Expected SpaceID or All.")

    def instances(
        self, action: Sequence[Action], spaces: set[str] | None = None, operation: str | None = None
    ) -> list[str] | None:
        """Validate access to data model instances.

        Args:
            action (Sequence[Action]): The actions to validate access for.
            spaces (Set[str] | None): The space IDs to check access for. If None, checks access for all spaces.
            operation (str | None): The operation being performed, used for error messages.

        Returns:
            list[str] | None: Returns a list of space IDs if access is limited to these spaces, or None if access is granted to all spaces.

        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected number of data model instance scopes.
            AuthorizationError: If the user does not have permission to perform the specified action on the given space.
        """
        operation = operation or self.default_operation
        instance_scopes = self._get_scopes(
            action, DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write, operation, "instances"
        )
        if len(instance_scopes) != 1:
            raise ValueError(
                f"Unexpected number of data model instance scopes: {len(instance_scopes)}. Expected 1 scope."
            )
        instance_scope = instance_scopes[0]
        if isinstance(instance_scope, DataModelInstancesAcl.Scope.SpaceID):
            if spaces is None:
                return instance_scope.space_ids
            if missing := spaces - set(instance_scope.space_ids):
                raise AuthorizationError(
                    f"You have no permission to {humanize_collection(action)} instances in the "
                    f"{humanize_collection(missing)!r} space(s). This is required to {operation} instances."
                )
            return None
        elif isinstance(instance_scope, DataModelInstancesAcl.Scope.All):
            return None
        else:
            raise ValueError(
                f"Unexpected data model instance scope type: {type(instance_scope)}. Expected SpaceID or All."
            )

    @overload
    def dataset_data(
        self,
        action: Sequence[Action],
        dataset_ids: set[int],
        operation: str | None = None,
        missing_access: Literal["raise", "warn"] = "raise",
    ) -> None: ...

    @overload
    def dataset_data(
        self,
        action: Sequence[Action],
        dataset_ids: None = None,
        operation: str | None = None,
        missing_access: Literal["raise", "warn"] = "raise",
    ) -> (
        dict[Literal["assets", "events", "time series", "files", "relationships", "labels", "3D models"], list[int]]
        | None
    ): ...

    def dataset_data(
        self,
        action: Sequence[Action],
        dataset_ids: set[int] | None = None,
        operation: str | None = None,
        missing_access: Literal["raise", "warn"] = "raise",
    ) -> (
        dict[Literal["assets", "events", "time series", "files", "relationships", "labels", "3D models"], list[int]]
        | None
    ):
        """Validate access to dataset data.

        Dataset data resources are:
        - Assets
        - Events
        - Time series
        - Files
        - Sequences
        - Relationships
        - Labels
        - 3D models

        Args:
            action (Sequence[Action]): The actions to validate access for
            dataset_ids (Set[int] | None): The dataset IDs to check access for. If None, checks access for all datasets.
            operation (str | None): The operation being performed, used for error and warning messages.
            missing_access (Literal["raise", "warn"]): Whether to raise an error or warn when access is missing for specified datasets.

        Returns:
            dict[
                Literal["assets", "events", "time series", "files", "relationships", "labels", "3D models"], list[int]
            ] | None:
                If dataset_ids is None, returns a dictionary with keys as dataset data resource names and values as lists of dataset IDs the user has access to.
                If dataset_ids is provided, returns None if the user has access to all specified datasets for all dataset data resources.
        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected dataset data scope type.
            AuthorizationError: If the user does not have permission to perform the specified action on the given dataset.
        """
        acls: list[tuple[str, list[Capability.Action], list[Capability.Action]]] = [
            ("assets", [AssetsAcl.Action.Read], [AssetsAcl.Action.Write]),
            ("events", [EventsAcl.Action.Read], [EventsAcl.Action.Write]),
            ("time series", [TimeSeriesAcl.Action.Read], [TimeSeriesAcl.Action.Write]),
            ("files", [FilesAcl.Action.Read], [FilesAcl.Action.Write]),
            ("sequences", [SequencesAcl.Action.Read], [SequencesAcl.Action.Write]),
            ("relationships", [RelationshipsAcl.Action.Read], [RelationshipsAcl.Action.Write]),
            ("labels", [LabelsAcl.Action.Read], [LabelsAcl.Action.Write]),
            (
                "3D models",
                [ThreeDAcl.Action.Read],
                [ThreeDAcl.Action.Create, ThreeDAcl.Action.Update, ThreeDAcl.Action.Delete],
            ),
        ]
        # MyPy does not understand that with the acl above, we get the correct return value.
        return self._dataset_access_check(  # type: ignore[return-value]
            action,
            dataset_ids=dataset_ids,
            operation=operation,
            acls=acls,
            missing_access=missing_access,
        )

    @overload
    def dataset_configurations(
        self,
        action: Sequence[Action],
        dataset_ids: set[int],
        operation: str | None = None,
        missing_access: Literal["raise", "warn"] = "raise",
    ) -> None: ...

    @overload
    def dataset_configurations(
        self,
        action: Sequence[Action],
        dataset_ids: None = None,
        operation: str | None = None,
        missing_access: Literal["raise", "warn"] = "raise",
    ) -> dict[Literal["transformations", "workflows", "extraction pipelines"], list[int]] | None: ...

    def dataset_configurations(
        self,
        action: Sequence[Action],
        dataset_ids: set[int] | None = None,
        operation: str | None = None,
        missing_access: Literal["raise", "warn"] = "raise",
    ) -> dict[Literal["transformations", "workflows", "extraction pipelines"], list[int]] | None:
        """Validate access configuration resources.

        Configuration resources are:
        - Transformations
        - Workflows
        - Extraction pipelines

        Args:
            action (Sequence[Action]): The actions to validate access for
            dataset_ids (Set[int] | None): The dataset IDs to check access for. If None, checks access for all datasets.
            operation (str | None): The operation being performed, used for error and warning messages.
            missing_access (Literal["raise", "warn"]): Whether to raise an error or warn when access is missing for specified datasets.

        Returns:
            dict[Literal["transformations", "workflows", "extraction pipelines"], list[int] | None]:
                If dataset_ids is None, returns a dictionary with keys as configuration resource names and values as lists of dataset IDs the user has access to.
                If dataset_ids is provided, returns None if the user has access to all specified datasets for all configuration resources.

        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected dataset configuration scope type.
            AuthorizationError: If the user does not have permission to perform the specified action on the given dataset.
        """
        acls: list[tuple[str, list[Capability.Action], list[Capability.Action]]] = [
            ("transformations", [TransformationsAcl.Action.Read], [TransformationsAcl.Action.Write]),
            ("workflows", [WorkflowOrchestrationAcl.Action.Read], [WorkflowOrchestrationAcl.Action.Write]),
            ("extraction pipelines", [ExtractionPipelinesAcl.Action.Read], [ExtractionPipelinesAcl.Action.Write]),
        ]
        # MyPy does not understand that with the acl above, we get the correct return value.
        return self._dataset_access_check(  # type: ignore[return-value]
            action,
            dataset_ids=dataset_ids,
            operation=operation,
            acls=acls,
            missing_access=missing_access,
        )

    def _dataset_access_check(
        self,
        action: Sequence[Action],
        dataset_ids: set[int] | None,
        operation: str | None,
        missing_access: Literal["raise", "warn"],
        acls: Sequence[tuple[str, list[Capability.Action], list[Capability.Action]]],
    ) -> dict[str, list[int]] | None:
        need_access_to = set(dataset_ids) if dataset_ids is not None else None
        no_access: list[str] = []
        output: dict[str, list[int]] = {}
        for name, read_actions, write_actions in acls:
            actions = [
                acl_action for word in action for acl_action in {"read": read_actions, "write": write_actions}[word]
            ]
            scopes = self.client.token.get_scope(actions)
            if scopes is None:
                no_access.append(name)
                continue
            # First check for 'all' scope
            for scope in scopes:
                if isinstance(scope, AllScope):
                    break
            else:
                # No 'all' scope found, check dataset scopes
                for scope in scopes:
                    if isinstance(scope, DataSetScope):
                        if need_access_to is None:
                            output[name] = scope.ids
                            break
                        missing_data_set = need_access_to - set(scope.ids)
                        if missing_data_set:
                            no_access.append(name)
                            break
        operation = operation or self.default_operation
        if no_access:
            message = f"You have no permission to {humanize_collection(action)} {humanize_collection(no_access)}."
            if dataset_ids:
                dataset_external_ids = self.client.lookup.data_sets.external_id(list(dataset_ids))
                plural = "s" if len(dataset_external_ids) > 1 else ""
                message = f"{message[:-1]} on dataset{plural} {humanize_collection(dataset_external_ids)}."
            if missing_access == "raise":
                raise AuthorizationError(f"{message} This is required to {operation}.")
            else:
                HighSeverityWarning(f"{message}. You will have limited functionality to {operation}.").print_warning()
        elif dataset_ids is not None:
            return None
        return output or None

    def timeseries(
        self,
        action: Sequence[Action],
        dataset_ids: set[int] | None = None,
        operation: str | None = None,
    ) -> dict[str, list[str]] | None:
        """Validate access to time series.
        Args:
            action (Sequence[Action]): The actions to validate access for.
            dataset_ids (Set[int] | None): The dataset IDs to check access for. If None, checks access for all datasets.
            operation (str | None): The operation being performed, used for error messages.
        Returns:
            dict[str, list[str]] | None: Returns a dictionary with keys 'dataset', 'asset root', and 'time series' if access is limited to these scopes, or None if access is granted to all time series.
        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected timeseries scope type.
            AuthorizationError: If the user does not have permission to perform the specified action on the given dataset or time series.
        """
        operation = operation or self.default_operation
        timeseries_scopes = self._get_scopes(
            action, TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write, operation, "time series"
        )

        if isinstance(timeseries_scopes[0], TimeSeriesAcl.Scope.All):
            return None
        if dataset_ids is not None:
            missing = set(dataset_ids)
            for scope in timeseries_scopes:
                if isinstance(scope, TimeSeriesAcl.Scope.DataSet):
                    missing = missing - set(scope.ids)
                    if not missing:
                        return None
            raise AuthorizationError(
                f"You have no permission to {humanize_collection(action)} time series in dataset {humanize_collection(missing)}. This is required to {operation}."
            )
        output: dict[str, list[str]] = {}
        for scope in timeseries_scopes:
            if isinstance(scope, TimeSeriesAcl.Scope.DataSet):
                output["dataset"] = self.client.lookup.data_sets.external_id(scope.ids)
            elif isinstance(scope, TimeSeriesAcl.Scope.AssetRootID):
                output["asset root"] = self.client.lookup.assets.external_id(scope.root_ids)
            elif isinstance(scope, TimeSeriesAcl.Scope.ID):
                output["time series"] = self.client.lookup.time_series.external_id(scope.ids)
            else:
                raise ValueError(
                    f"Unexpected timeseries scope type: {type(scope)}. Expected DataSet, AssetRootID or ID."
                )

        return output

    def files(
        self,
        action: Sequence[Action],
        dataset_ids: set[int] | None = None,
        operation: str | None = None,
    ) -> dict[str, list[str]] | None:
        """Validate access to files.

        Args:
            action (Sequence[Action]): The actions to validate access for
            dataset_ids (Set[int] | None): The dataset IDs to check access for. If None, checks access for all datasets.
            operation (str | None): The operation being performed, used for error messages.
        Returns:
            dict[str, list[str]] | None: Returns a dictionary with the key 'dataset' if access is limited to a dataset scope, or None if access is granted to all files.
        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected file scope type.
            AuthorizationError: If the user does not have permission to perform the specified action on the given
                dataset.
        """
        operation = operation or self.default_operation
        file_scopes = self._get_scopes(action, FilesAcl.Action.Read, FilesAcl.Action.Write, operation, "files")
        if isinstance(file_scopes[0], FilesAcl.Scope.All):
            return None
        if dataset_ids is not None:
            missing = set(dataset_ids)
            for scope in file_scopes:
                if isinstance(scope, FilesAcl.Scope.DataSet):
                    missing = missing - set(scope.ids)
                    if not missing:
                        return None
            raise AuthorizationError(
                f"You have no permission to {humanize_collection(action)} files in dataset "
                f"{humanize_collection(missing)}. This is required to {operation}."
            )
        output: dict[str, list[str]] = {}
        for scope in file_scopes:
            if isinstance(scope, FilesAcl.Scope.DataSet):
                output["dataset"] = self.client.lookup.data_sets.external_id(scope.ids)
            else:
                raise ValueError(f"Unexpected file scope type: {type(scope)}. Expected DataSet or All.")
        return output

    def assets(
        self,
        action: Sequence[Action],
        dataset_ids: set[int] | None = None,
        operation: str | None = None,
    ) -> dict[str, list[str]] | None:
        """Validate access to assets.
        Args:
            action (Sequence[Action]): The actions to validate access for
            dataset_ids (Set[int] | None): The dataset IDs to check access for. If None, checks access for all datasets.
            operation (str | None): The operation being performed, used for error messages.
        Returns:
            dict[str, list[str]] | None: Returns a dictionary with the key 'dataset'
                if access is limited to a dataset scope, or None if access is granted to all assets.
        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected asset scope type.
            AuthorizationError: If the user does not have permission to perform the specified action on the given
                dataset.
        """
        operation = operation or self.default_operation
        asset_scopes = self._get_scopes(action, AssetsAcl.Action.Read, AssetsAcl.Action.Write, operation, "assets")
        if isinstance(asset_scopes[0], AssetsAcl.Scope.All):
            return None
        if dataset_ids is not None:
            missing = set(dataset_ids)
            for scope in asset_scopes:
                if isinstance(scope, AssetsAcl.Scope.DataSet):
                    missing = missing - set(scope.ids)
                    if not missing:
                        return None
            raise AuthorizationError(
                f"You have no permission to {humanize_collection(action)} assets in dataset(s) "
                f"{humanize_collection(missing)}. This is required to {operation}."
            )
        output: dict[str, list[str]] = {}
        for scope in asset_scopes:
            if isinstance(scope, AssetsAcl.Scope.DataSet):
                output["dataset"] = self.client.lookup.data_sets.external_id(scope.ids)
            else:
                raise ValueError(f"Unexpected asset scope type: {type(scope)}. Expected DataSet or All.")
        return output

    def _get_scopes(
        self,
        action: Sequence[Action],
        read: Capability.Action,
        write: Capability.Action,
        operation: str,
        name: str,
    ) -> list[Capability.Scope]:
        """Helper method to get scopes for the given action.

        Args:
            action (Sequence[Action]): The actions to validate access for.
            read (Capability.Action): The read action.
            write (Capability.Action): The write action.
            operation (str): The operation being performed, used for error messages.
            name (str): The name of the resource being accessed, used for error messages.

        Returns:
            list[Capability.Scope]: The scopes for the given action.
        """
        actions = [{"read": read, "write": write}[a] for a in action]
        scopes = self.client.token.get_scope(actions)
        if scopes is None:
            raise AuthorizationError(
                f"You have no permission to {humanize_collection(action, bind_word='and')} {name}. "
                f"This is required to {operation}."
            )
        return scopes



from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import Literal

from cognite.client.data_classes.data_modeling import ViewId

from cognite_toolkit._cdf_tk.commands._migrate.data_classes import MigrationMappingList
from cognite_toolkit._cdf_tk.storageio import DataSelector
from cognite_toolkit._cdf_tk.storageio.selectors import DataSetSelector
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricKind


class AssetCentricMigrationSelector(DataSelector, ABC):
    kind: AssetCentricKind

    @abstractmethod
    def get_ingestion_mappings(self) -> list[str]:
        raise NotImplementedError()


class MigrationCSVFileSelector(AssetCentricMigrationSelector):
    type: Literal["migrationCSVFile"] = "migrationCSVFile"
    datafile: Path

    @property
    def group(self) -> str:
        return f"Migration_{self.kind}"

    def __str__(self) -> str:
        return f"file_{self.datafile.name}"

    def get_ingestion_mappings(self) -> list[str]:
        views = {item.get_ingestion_view() for item in self.items}
        return sorted(views)

    @cached_property
    def items(self) -> MigrationMappingList:
        return MigrationMappingList.read_csv_file(self.datafile, resource_type=self.kind)


class MigrateDataSetSelector(AssetCentricMigrationSelector):
    type: Literal["migrateDataSet"] = "migrateDataSet"
    kind: AssetCentricKind
    data_set_external_id: str
    ingestion_mapping: str | None = None
    preferred_consumer_view: ViewId | None = None

    @property
    def group(self) -> str:
        return f"DataSet_{self.data_set_external_id}"

    def __str__(self) -> str:
        return self.kind

    def get_schema_spaces(self) -> list[str] | None:
        return None

    def get_instance_spaces(self) -> list[str] | None:
        return None

    def get_ingestion_mappings(self) -> list[str]:
        return [self.ingestion_mapping] if self.ingestion_mapping else []

    def as_asset_centric_selector(self) -> DataSetSelector:
        return DataSetSelector(data_set_external_id=self.data_set_external_id, kind=self.kind)



import json
from collections.abc import Iterator
from pathlib import Path
from unittest.mock import MagicMock

import httpx
import pytest
import responses
import respx
from cognite.client.data_classes import Asset, AssetList
from cognite.client.data_classes.data_modeling import (
    ContainerId,
    DataModel,
    DataModelList,
    DirectRelation,
    MappedProperty,
    NodeApply,
    NodeOrEdgeData,
    Text,
    View,
    ViewId,
)
from cognite.client.data_classes.data_modeling.statistics import InstanceStatistics, ProjectStatistics

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands._migrate.command import MigrationCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import AssetCentricMapper
from cognite_toolkit._cdf_tk.commands._migrate.data_model import (
    COGNITE_MIGRATION_MODEL,
    INSTANCE_SOURCE_VIEW_ID,
    MODEL_ID,
    RESOURCE_VIEW_MAPPING_VIEW_ID,
)
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import ASSET_ID, create_default_mappings
from cognite_toolkit._cdf_tk.commands._migrate.migration_io import AssetCentricMigrationIO
from cognite_toolkit._cdf_tk.commands._migrate.selectors import MigrationCSVFileSelector
from cognite_toolkit._cdf_tk.exceptions import ToolkitMigrationError, ToolkitValueError
from cognite_toolkit._cdf_tk.utils.fileio import CSVReader


@pytest.fixture
def cognite_migration_model(
    toolkit_config: ToolkitClientConfig, rsps: responses.RequestsMock
) -> Iterator[responses.RequestsMock]:
    config = toolkit_config
    mapping_by_id = {mapping.external_id: mapping for mapping in create_default_mappings()}
    asset_mapping = mapping_by_id[ASSET_ID]
    # Lookup of the mapping in the Migration Model
    mapping_node_response = asset_mapping.dump(context="api")
    mapping_node_response.update({"createdTime": 0, "lastUpdatedTime": 0, "version": 1})
    sources = mapping_node_response.pop("sources", [])
    if sources:
        mapping_view_id = asset_mapping.sources[0].source
        mapping_node_response["properties"] = {
            mapping_view_id.space: {
                f"{mapping_view_id.external_id}/{mapping_view_id.version}": sources[0]["properties"]
            }
        }
    rsps.post(
        config.create_api_url("models/instances/byids"),
        json={"items": [mapping_node_response]},
        status=200,
    )

    # Lookup CogniteAsset, this is not the full model, just the properties we need for the
    # migration
    default_prop_args = dict(nullable=True, immutable=False, auto_increment=False)
    default_view_args = dict(
        last_updated_time=1,
        created_time=1,
        description=None,
        name=None,
        implements=[],
        writable=True,
        used_for="node",
        is_global=True,
        filter=None,
    )
    cognite_asset = View(
        space="cdf_cdm",
        external_id="CogniteAsset",
        version="v1",
        properties={
            "name": MappedProperty(ContainerId("cdf_cdm", "CogniteDescribable"), "name", Text(), **default_prop_args),
            "description": MappedProperty(
                ContainerId("cdf_cdm", "CogniteDescribable"), "description", Text(), **default_prop_args
            ),
            "source": MappedProperty(
                ContainerId("cdf_cdm", "CogniteSourceable"), "name", DirectRelation(), **default_prop_args
            ),
            "tags": MappedProperty(
                ContainerId("cdf_cdm", "CogniteDescribable"), "tags", Text(is_list=True), **default_prop_args
            ),
        },
        **default_view_args,
    )
    rsps.post(
        config.create_api_url("models/views/byids"),
        json={"items": [cognite_asset.dump()]},
    )
    # Migration model
    migration_model = COGNITE_MIGRATION_MODEL.dump()
    migration_model["createdTime"] = 1
    migration_model["lastUpdatedTime"] = 1
    migration_model["isGlobal"] = False
    rsps.post(config.create_api_url("models/datamodels/byids"), json={"items": migration_model})

    yield rsps


@pytest.fixture
def mock_statistics(
    toolkit_config: ToolkitClientConfig,
    rsps: responses.RequestsMock,
) -> Iterator[responses.RequestsMock]:
    config = toolkit_config
    stats_response = {
        "spaces": {
            "count": 0,
            "limit": 1_000,
        },
        "containers": {
            "count": 0,
            "limit": 10_000,
        },
        "views": {
            "count": 0,
            "limit": 100_000,
        },
        "dataModels": {
            "count": 1,
            "limit": 10_000,
        },
        "containerProperties": {
            "count": 0,
            "limit": 1_000_000,
        },
        "instances": {
            "nodes": 1000,
            "edges": 0,
            "softDeletedNodes": 0,
            "softDeletedEdges": 0,
            "instancesLimit": 5_000_000,
            "softDeletedInstancesLimit": 100_000_000,
            "instances": 1000,
            "softDeletedInstances": 0,
        },
        "concurrentReadLimit": 50,
        "concurrentWriteLimit": 20,
        "concurrentDeleteLimit": 10,
    }
    rsps.get(
        config.create_api_url("/models/statistics"),
        json=stats_response,
        status=200,
    )
    yield rsps


class TestMigrationCommand:
    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check", "mock_statistics")
    def test_migrate_assets(
        self,
        toolkit_config: ToolkitClientConfig,
        cognite_migration_model: responses.RequestsMock,
        tmp_path: Path,
        respx_mock: respx.MockRouter,
    ) -> None:
        rsps = cognite_migration_model
        config = toolkit_config
        assets = AssetList(
            [
                Asset(
                    id=1000 + i,
                    external_id=f"asset_{i}",
                    name=f"Asset {i}",
                    description=f"This is Asset {i}",
                    last_updated_time=1,
                    created_time=0,
                    parent_external_id="asset_0" if i > 0 else None,
                )
                for i in range(2)
            ]
        )
        space = "my_space"
        csv_content = "id,space,externalId,ingestionView\n" + "\n".join(
            f"{1000 + i},{space},asset_{i},{ASSET_ID}" for i in range(len(assets))
        )

        # Asset retrieve ids
        rsps.post(
            config.create_api_url("/assets/byids"),
            json={"items": [asset.dump() for asset in assets]},
            status=200,
        )
        # Lookup CogniteSourceSystem (no source systems defined)
        rsps.post(
            config.create_api_url("/models/instances/list"),
            json={"items": []},
            status=200,
        )
        # Instance creation
        respx.post(
            config.create_api_url("/models/instances"),
        ).mock(
            return_value=httpx.Response(
                status_code=200,
                json={
                    "items": [
                        {
                            "instanceType": "node",
                            "space": space,
                            "externalId": f"asset_{i}",
                            "version": 1,
                            "wasModified": True,
                            "createdTime": 1,
                            "lastUpdatedTime": 1,
                        }
                        for i in range(len(assets))
                    ]
                },
            )
        )
        csv_file = tmp_path / "migration.csv"
        csv_file.write_text(csv_content, encoding="utf-8")

        client = ToolkitClient(config)
        command = MigrationCommand(silent=True)

        result = command.migrate(
            selected=MigrationCSVFileSelector(datafile=csv_file, kind="Assets"),
            data=AssetCentricMigrationIO(client),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path / "logs",
            dry_run=False,
            verbose=False,
        )

        # Check that the assets were uploaded
        last_call = respx_mock.calls[-1]
        assert last_call.request.url == config.create_api_url("/models/instances")
        assert last_call.request.method == "POST"
        actual_instances = json.loads(last_call.request.content)["items"]
        expected_instance = [
            NodeApply(
                space=space,
                external_id=asset.external_id,
                sources=[
                    NodeOrEdgeData(
                        source=ViewId("cdf_cdm", "CogniteAsset", "v1"),
                        properties={
                            "name": asset.name,
                            "description": asset.description,
                        },
                    ),
                    NodeOrEdgeData(
                        source=INSTANCE_SOURCE_VIEW_ID,
                        properties={
                            "id": asset.id,
                            "resourceType": "asset",
                            "dataSetId": None,
                            "classicExternalId": asset.external_id,
                        },
                    ),
                ],
            ).dump()
            for asset in assets
        ]
        assert actual_instances == expected_instance
        actual_results = [result.get_progress(f"asset_{asset.id}") for asset in assets]
        expected_results = [{"download": "success", "convert": "success", "upload": "success"} for _ in assets]
        assert actual_results == expected_results
        csv_file = next((tmp_path / "logs").glob("*.csv"), None)
        assert csv_file is not None, "Expected a CSV log file to be created"
        csv_results = list(CSVReader(csv_file).read_chunks_unprocessed())
        assert csv_results == [
            {"ID": f"asset_{asset.id}", "download": "success", "convert": "success", "upload": "success"}
            for asset in assets
        ]

    def test_validate_migration_model_available(self) -> None:
        with monkeypatch_toolkit_client() as client:
            client.data_modeling.data_models.retrieve.return_value = DataModelList([])
            with pytest.raises(ToolkitMigrationError):
                MigrationCommand.validate_migration_model_available(client)

    def test_validate_migration_model_available_multiple_models(self) -> None:
        """Test that multiple models raises an error."""
        with monkeypatch_toolkit_client() as client:
            # Create mock models with the expected MODEL_ID
            model1 = MagicMock(spec=DataModel)
            model1.as_id.return_value = MODEL_ID
            model2 = MagicMock(spec=DataModel)
            model2.as_id.return_value = MODEL_ID

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model1, model2])

            with pytest.raises(ToolkitMigrationError) as exc_info:
                MigrationCommand.validate_migration_model_available(client)

            assert "Multiple migration models" in str(exc_info.value)

    def test_validate_migration_model_available_missing_views(self) -> None:
        """Test that a model with missing views raises an error."""
        with monkeypatch_toolkit_client() as client:
            model = MagicMock(spec=DataModel)
            model.as_id.return_value = MODEL_ID
            # Model has views but missing the required ones
            model.views = [INSTANCE_SOURCE_VIEW_ID]  # Missing VIEW_SOURCE_VIEW_ID

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model])

            with pytest.raises(ToolkitMigrationError, match=r"Invalid migration model. Missing views"):
                MigrationCommand.validate_migration_model_available(client)

    def test_validate_migration_model_available_success(self) -> None:
        """Test that a valid model with all required views succeeds."""
        with monkeypatch_toolkit_client() as client:
            # Mocking the migration Model to get a response format of the model.
            # An alternative would be to write a conversion of write -> read format of the model
            # which is a significant amount of logic.
            model = MagicMock(spec=DataModel)
            model.as_id.return_value = MODEL_ID
            # Model has all required views
            model.views = [INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID]

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model])

            # Should not raise any exception
            MigrationCommand.validate_migration_model_available(client)

            client.data_modeling.data_models.retrieve.assert_called_once_with([MODEL_ID], inline_views=False)

    def test_validate_available_capacity_missing_capacity(self) -> None:
        cmd = MigrationCommand(silent=True)

        with monkeypatch_toolkit_client() as client:
            stats = MagicMock(spec=ProjectStatistics)
            stats.instances = InstanceStatistics(
                nodes=1000,
                edges=0,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
                instances_limit=1500,
                soft_deleted_instances_limit=10_000,
                instances=1000,
                soft_deleted_instances=0,
            )
            client.data_modeling.statistics.project.return_value = stats
            with pytest.raises(ToolkitValueError) as exc_info:
                cmd.validate_available_capacity(client, 10_000)

        assert "Cannot proceed with migration" in str(exc_info.value)

    def test_validate_available_capacity_sufficient_capacity(self) -> None:
        cmd = MigrationCommand(silent=True)

        with monkeypatch_toolkit_client() as client:
            stats = MagicMock(spec=ProjectStatistics)
            stats.instances = InstanceStatistics(
                nodes=1000,
                edges=0,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
                instances_limit=5_000_000,
                soft_deleted_instances_limit=100_000_000,
                instances=1000,
                soft_deleted_instances=0,
            )
            client.data_modeling.statistics.project.return_value = stats
            cmd.validate_available_capacity(client, 10_000)



from typing import Literal

from cognite.client.data_classes import WorkflowVersionUpsert
from pydantic import Field, JsonValue

from .base import BaseModelResource, ToolkitResource


class TaskId(BaseModelResource):
    external_id: str = Field(
        max_length=255, description="The external ID provided by the client. Must be unique for the resource type."
    )


class TaskDefinition(BaseModelResource):
    external_id: str = Field(
        max_length=255,
        description="Identifier for the task. Must be unique within the version. No trailing or leading whitespace and no null characters allowed.",
    )
    type: Literal["function", "transformation", "cdf", "dynamic", "subworkflow", "simulation"]
    name: str | None = Field(
        default=None,
        max_length=255,
        description="Readable name meant for use in UIs",
    )
    description: str | None = Field(
        default=None,
        max_length=500,
        description="Description of the intention of the task",
    )
    parameters: JsonValue = Field()
    retries: int = Field(
        3,
        ge=0,
        le=10,
        description="Number of times to retry the task if it fails. If set to 0, the task will not be retried. The behavior for timeouts and retries is defined by the onFailure parameter, refer to it for more information.",
    )
    timeout: int = Field(
        3600,
        ge=60,
        le=86400,
        description="TTimeout in seconds. After this time, the task will be marked as TIMED_OUT. By default, the task won't be retried upon timeout. Use the onFailure parameter to change this behavior.",
    )
    on_failure: Literal["abortWorkflow", "skipTask"] = Field(
        "abortWorkflow",
        description="Defines the behavior when a task fails. If set to abortWorkflow, the entire workflow will be marked as FAILED when the task fails. If set to skipTask, the task will be marked as FAILED, but the workflow will continue executing subsequent tasks that are not dependent on the failed task.",
    )
    depends_on: list[TaskId] | None = Field(
        None,
        description="The tasks that must be completed before this task can be executed.",
        max_length=100,
    )


class WorkflowDefinition(BaseModelResource):
    description: str | None = Field(
        default=None,
        max_length=500,
        description="The description of the workflow version.",
    )
    tasks: list[TaskDefinition]


class WorkflowVersionYAML(ToolkitResource):
    _cdf_resource = WorkflowVersionUpsert
    workflow_external_id: str = Field(
        max_length=255,
        description="Identifier for a workflow. Must be unique for the project. No trailing or leading whitespace and no null characters allowed.",
    )
    version: str = Field(
        max_length=255,
        description="Identifier for a version. Must be unique for the workflow. No trailing or leading whitespace and no null characters allowed.",
    )
    workflow_definition: WorkflowDefinition



from collections.abc import Iterator, Sequence
from typing import overload

from cognite.client._api_client import APIClient
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.robotics import (
    DataPostProcessing,
    DataPostProcessingList,
    DataPostProcessingWrite,
    _DataProcessingUpdate,
)

from .utlis import tmp_disable_gzip


class DataPostProcessingAPI(APIClient):
    _RESOURCE_PATH = "/robotics/data_postprocessing"

    @overload
    def __call__(self) -> Iterator[DataPostProcessing]: ...

    @overload
    def __call__(self, chunk_size: int) -> Iterator[DataPostProcessingList]: ...

    def __call__(
        self, chunk_size: int | None = None
    ) -> Iterator[DataPostProcessing] | Iterator[DataPostProcessingList]:
        """Iterate over robot data_processing.

        Args:
            chunk_size: The number of robot data_processing to return in each chunk. None will return all robot data_processing.

        Yields:
            DataProcessing or DataProcessingList

        """
        return self._list_generator(
            method="GET", resource_cls=DataPostProcessing, list_cls=DataPostProcessingList, chunk_size=chunk_size
        )

    def __iter__(self) -> Iterator[DataPostProcessing]:
        return self.__call__()

    @overload
    def create(self, dataProcessing: DataPostProcessingWrite) -> DataPostProcessing: ...

    @overload
    def create(self, dataProcessing: Sequence[DataPostProcessingWrite]) -> DataPostProcessingList: ...

    def create(
        self, dataProcessing: DataPostProcessingWrite | Sequence[DataPostProcessingWrite]
    ) -> DataPostProcessing | DataPostProcessingList:
        """Create a new robot dataProcessing.

        Args:
            dataProcessing: DataProcessingWrite or list of DataProcessingWrite.

        Returns:
            DataProcessing object.

        """
        with tmp_disable_gzip():
            return self._create_multiple(
                list_cls=DataPostProcessingList,
                resource_cls=DataPostProcessing,
                items=dataProcessing,
                input_resource_cls=DataPostProcessingWrite,
            )

    @overload
    def retrieve(self, external_id: str) -> DataPostProcessing | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str]) -> DataPostProcessingList: ...

    def retrieve(self, external_id: str | SequenceNotStr[str]) -> DataPostProcessing | None | DataPostProcessingList:
        """Retrieve a robot dataProcessing.

        Args:
            external_id: External id of the robot dataProcessing.

        Returns:
            DataProcessing object.

        """
        identifiers = IdentifierSequence.load(external_ids=external_id)
        with tmp_disable_gzip():
            return self._retrieve_multiple(
                identifiers=identifiers,
                resource_cls=DataPostProcessing,
                list_cls=DataPostProcessingList,
            )

    @overload
    def update(self, dataProcessing: DataPostProcessingWrite) -> DataPostProcessing: ...

    @overload
    def update(self, dataProcessing: Sequence[DataPostProcessingWrite]) -> DataPostProcessingList: ...

    def update(
        self, dataProcessing: DataPostProcessingWrite | Sequence[DataPostProcessingWrite]
    ) -> DataPostProcessing | DataPostProcessingList:
        """Update a robot dataProcessing.

        Args:
            dataProcessing: DataProcessingWrite or list of DataProcessingWrite.

        Returns:
            DataProcessing object.

        """
        with tmp_disable_gzip():
            return self._update_multiple(
                items=dataProcessing,
                resource_cls=DataPostProcessing,
                list_cls=DataPostProcessingList,
                update_cls=_DataProcessingUpdate,
            )

    def delete(self, external_id: str | SequenceNotStr[str]) -> None:
        """Delete a robot dataProcessing.

        Args:
            external_id: External id of the robot dataProcessing.

        Returns:
            None

        """
        identifiers = IdentifierSequence.load(external_ids=external_id)
        with tmp_disable_gzip():
            self._delete_multiple(identifiers=identifiers, wrap_ids=True)

    def list(self) -> DataPostProcessingList:
        """List robot data_processing.

        Returns:
            DataProcessingList

        """
        with tmp_disable_gzip():
            return self._list(method="GET", resource_cls=DataPostProcessing, list_cls=DataPostProcessingList)



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
from collections.abc import Hashable, Iterable, Sequence
from graphlib import CycleError, TopologicalSorter
from pathlib import Path
from typing import Any, final

from cognite.client.data_classes import (
    ClientCredentials,
    Workflow,
    WorkflowList,
    WorkflowTrigger,
    WorkflowTriggerList,
    WorkflowTriggerUpsert,
    WorkflowTriggerUpsertList,
    WorkflowUpsert,
    WorkflowUpsertList,
    WorkflowVersion,
    WorkflowVersionId,
    WorkflowVersionList,
    WorkflowVersionUpsert,
    WorkflowVersionUpsertList,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    WorkflowOrchestrationAcl,
)
from cognite.client.data_classes.workflows import SubworkflowReferenceParameters
from cognite.client.exceptions import CogniteAuthError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceCreationError,
    ToolkitCycleError,
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.resource_classes import WorkflowTriggerYAML, WorkflowVersionYAML, WorkflowYAML
from cognite_toolkit._cdf_tk.tk_warnings import (
    LowSeverityWarning,
    MissingReferencedWarning,
    ToolkitWarning,
)
from cognite_toolkit._cdf_tk.utils import (
    calculate_secure_hash,
    humanize_collection,
    load_yaml_inject_variables,
    sanitize_filename,
)
from cognite_toolkit._cdf_tk.utils.cdf import read_auth, try_find_error
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable

from .auth import GroupAllScopedCRUD
from .data_organization import DataSetsCRUD
from .function import FunctionCRUD
from .group_scoped import GroupResourceScopedCRUD
from .transformation import TransformationCRUD


@final
class WorkflowCRUD(ResourceCRUD[str, WorkflowUpsert, Workflow, WorkflowUpsertList, WorkflowList]):
    folder_name = "workflows"
    filename_pattern = r"^.*Workflow$"
    resource_cls = Workflow
    resource_write_cls = WorkflowUpsert
    list_cls = WorkflowList
    list_write_cls = WorkflowUpsertList
    kind = "Workflow"
    dependencies = frozenset(
        {
            GroupAllScopedCRUD,
            TransformationCRUD,
            FunctionCRUD,
            DataSetsCRUD,
        }
    )
    yaml_cls = WorkflowYAML
    _doc_base_url = "https://api-docs.cognite.com/20230101-beta/tag/"
    _doc_url = "Workflows/operation/CreateOrUpdateWorkflow"

    @property
    def display_name(self) -> str:
        return "workflows"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[WorkflowUpsert] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [WorkflowOrchestrationAcl.Action.Read]
            if read_only
            else [WorkflowOrchestrationAcl.Action.Read, WorkflowOrchestrationAcl.Action.Write]
        )

        return WorkflowOrchestrationAcl(
            actions,
            WorkflowOrchestrationAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: Workflow | WorkflowUpsert | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("Workflow must have external_id set.")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> WorkflowUpsert:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return WorkflowUpsert._load(resource)

    def dump_resource(self, resource: Workflow, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        return dumped

    def retrieve(self, ids: SequenceNotStr[str]) -> WorkflowList:
        return self.client.workflows.retrieve(external_id=ids, ignore_unknown_ids=True)

    def _upsert(self, items: WorkflowUpsert | WorkflowUpsertList) -> WorkflowList:
        upserts = [items] if isinstance(items, WorkflowUpsert) else items
        return self.client.workflows.upsert(upserts)

    def create(self, items: WorkflowUpsert | WorkflowUpsertList) -> WorkflowList:
        return self._upsert(items)

    def update(self, items: WorkflowUpsertList) -> WorkflowList:
        return self._upsert(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        successes = 0
        for id_ in ids:
            try:
                self.client.workflows.delete(external_id=id_)
            except CogniteNotFoundError:
                print(f"  [bold yellow]WARNING:[/] Workflow {id_} does not exist, skipping delete.")
            else:
                successes += 1
        return successes

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Workflow]:
        if data_set_external_id is None:
            yield from self.client.workflows.list(limit=-1)
            return
        data_set = self.client.data_sets.retrieve(external_id=data_set_external_id)
        if data_set is None:
            raise ToolkitRequiredValueError(f"DataSet {data_set_external_id!r} does not exist")
        for workflow in self.client.workflows.list(limit=-1):
            if workflow.data_set_id == data_set.id:
                yield workflow

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]


@final
class WorkflowVersionCRUD(
    ResourceCRUD[
        WorkflowVersionId, WorkflowVersionUpsert, WorkflowVersion, WorkflowVersionUpsertList, WorkflowVersionList
    ]
):
    folder_name = "workflows"
    filename_pattern = r"^.*WorkflowVersion$"
    resource_cls = WorkflowVersion
    resource_write_cls = WorkflowVersionUpsert
    list_cls = WorkflowVersionList
    list_write_cls = WorkflowVersionUpsertList
    kind = "WorkflowVersion"
    dependencies = frozenset({WorkflowCRUD})
    parent_resource = frozenset({WorkflowCRUD})
    yaml_cls = WorkflowVersionYAML

    _doc_base_url = "https://api-docs.cognite.com/20230101-beta/tag/"
    _doc_url = "Workflow-versions/operation/CreateOrUpdateWorkflowVersion"
    # The /list endpoint has a limit of 100 workflow ids per request.
    _list_workflow_filter_limit = 100

    @property
    def display_name(self) -> str:
        return "workflow versions"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[WorkflowVersionUpsert] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [WorkflowOrchestrationAcl.Action.Read]
            if read_only
            else [WorkflowOrchestrationAcl.Action.Read, WorkflowOrchestrationAcl.Action.Write]
        )

        return WorkflowOrchestrationAcl(
            actions,
            WorkflowOrchestrationAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: WorkflowVersion | WorkflowVersionUpsert | dict) -> WorkflowVersionId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"workflowExternalId", "version"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return WorkflowVersionId(item["workflowExternalId"], item["version"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: WorkflowVersionId) -> dict[str, Any]:
        return id.dump()

    def load_resource_file(
        self,
        filepath: Path,
        environment_variables: dict[str, str | None] | None = None,
    ) -> list[dict[str, Any]]:
        # Special case where the environment variable keys in the 'workflowDefinition.tasks.parameters' are
        # JSONPaths that are part of the API and thus should not be replaced.
        raw_str = self.safe_read(filepath)

        original = load_yaml_inject_variables(raw_str, {}, validate=False, original_filepath=filepath)
        replaced = load_yaml_inject_variables(
            raw_str, environment_variables or {}, validate=False, original_filepath=filepath
        )

        if isinstance(original, dict) and isinstance(replaced, dict):
            self._update_task_parameters(replaced, original)
            return [replaced]
        elif isinstance(original, list) and isinstance(replaced, list):
            for original_item, replaced_item in zip(original, replaced):
                self._update_task_parameters(replaced_item, original_item)
            return replaced
        else:
            # Should be unreachable
            raise ValueError(
                f"Unexpected state. Loaded {filepath.as_posix()!r} twice and got different types: {type(original)} and {type(replaced)}"
            )

    @staticmethod
    def _update_task_parameters(replaced: dict, original: dict) -> None:
        replaced_def = replaced.get("workflowDefinition")
        original_def = original.get("workflowDefinition")
        if not (isinstance(replaced_def, dict) and isinstance(original_def, dict)):
            return
        replaced_tasks = replaced_def.get("tasks")
        original_tasks = original_def.get("tasks")
        if not (isinstance(replaced_tasks, list) and isinstance(original_tasks, list)):
            return
        for replaced_task, original_task in zip(replaced_def["tasks"], original_def["tasks"]):
            if not (isinstance(replaced_task, dict) and isinstance(original_task, dict)):
                continue
            if "parameters" in replaced_task and "parameters" in original_task:
                replaced_task["parameters"] = original_task["parameters"]

    def dump_resource(self, resource: WorkflowVersion, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        if not local:
            return dumped
        # Sort to match the order of the local tasks
        if "workflowDefinition" in local and "tasks" in local["workflowDefinition"]:
            local_task_order_by_id = {
                task["externalId"]: no for no, task in enumerate(local["workflowDefinition"]["tasks"])
            }
            local_task_by_id = {task["externalId"]: task for task in local["workflowDefinition"]["tasks"]}
        else:
            local_task_order_by_id = {}
            local_task_by_id = {}
        end_of_list = len(local_task_order_by_id)
        dumped["workflowDefinition"]["tasks"] = sorted(
            dumped["workflowDefinition"]["tasks"],
            key=lambda t: local_task_order_by_id.get(t["externalId"], end_of_list),
        )

        # Function tasks with empty data can be an empty dict or missing data field
        # This ensures that these two are treated as the same
        for cdf_task in dumped["workflowDefinition"]["tasks"]:
            task_id = cdf_task["externalId"]
            if task_id not in local_task_by_id:
                continue
            local_task = local_task_by_id[task_id]
            for key, default_value in [("retries", 3), ("timeout", 3600), ("onFailure", "abortWorkflow")]:
                if key not in local_task and cdf_task.get(key) == default_value:
                    del cdf_task[key]
                elif (
                    key in local_task
                    and local_task[key] is None
                    and (cdf_task.get(key) == default_value or key not in cdf_task)
                ):
                    cdf_task[key] = None

            if local_task["type"] == "function" and cdf_task["type"] == "function":
                cdf_parameters = cdf_task["parameters"]
                local_parameters = local_task["parameters"]
                if "function" in cdf_parameters and "function" in local_parameters:
                    cdf_function = cdf_parameters["function"]
                    local_function = local_parameters["function"]
                    if local_function.get("data") == {} and "data" not in cdf_function:
                        cdf_parameters["function"] = local_function
            elif local_task["type"] == "transformation" and cdf_task["type"] == "transformation":
                cdf_parameters = cdf_task["parameters"]
                local_parameters = local_task["parameters"]
                if "transformation" in cdf_parameters and "transformation" in local_parameters:
                    for key, default_transformation_value in [
                        ("concurrencyPolicy", "fail"),
                        ("useTransformationCredentials", False),
                    ]:
                        if (
                            key not in local_parameters["transformation"]
                            and cdf_parameters["transformation"].get(key) == default_transformation_value
                        ):
                            del cdf_parameters["transformation"][key]
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("workflowDefinition", "tasks"):
            return diff_list_identifiable(local, cdf, get_identifier=lambda t: t["externalId"])
        elif len(json_path) == 4 and json_path[:2] == ("workflowDefinition", "tasks") and json_path[3] == "dependsOn":
            return diff_list_identifiable(local, cdf, get_identifier=lambda t: t["externalId"])
        elif len(json_path) >= 2 and json_path[:2] == ("workflowDefinition", "tasks"):
            # Assume all other arrays in the tasks are hashable
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "workflowExternalId" in item:
            yield WorkflowCRUD, item["workflowExternalId"]

    @classmethod
    def check_item(cls, item: dict, filepath: Path, element_no: int | None) -> list[ToolkitWarning]:
        warnings: list[ToolkitWarning] = []
        tasks = item.get("workflowDefinition", {}).get("tasks", [])
        if not tasks:
            # We do not check for tasks here, that is done by comparing the spec.
            return warnings
        # Checking for invalid dependsOn
        tasks_ids = {task["externalId"] for task in tasks if "externalId" in task}
        for task in tasks:
            if not isinstance(depends_on := task.get("dependsOn"), list):
                continue
            invalid_tasks = [
                dep["externalId"] for dep in depends_on if "externalId" in dep and dep["externalId"] not in tasks_ids
            ]
            if invalid_tasks:
                warnings.append(
                    MissingReferencedWarning(
                        filepath=filepath,
                        element_no=element_no,
                        path=tuple(),
                        message=f"Task {task['externalId']!r} depends on non-existing task(s): {humanize_collection(invalid_tasks)!r}",
                    )
                )
        return warnings

    def retrieve(self, ids: SequenceNotStr[WorkflowVersionId]) -> WorkflowVersionList:
        if not ids:
            return WorkflowVersionList([])
        return self.client.workflows.versions.retrieve(workflow_external_id=list(ids), ignore_unknown_ids=True)

    def _upsert(self, items: WorkflowVersionUpsertList) -> WorkflowVersionList:
        return self.client.workflows.versions.upsert(items)

    def create(self, items: WorkflowVersionUpsertList) -> WorkflowVersionList:
        upserted: list[WorkflowVersion] = []
        for item in self.topological_sort(items):
            upserted.extend(self.client.workflows.versions.upsert([item]))
        return WorkflowVersionList(upserted)

    def update(self, items: WorkflowVersionUpsertList) -> WorkflowVersionList:
        return self._upsert(items)

    def delete(self, ids: SequenceNotStr[WorkflowVersionId]) -> int:
        successes = 0
        for id in ids:
            try:
                self.client.workflows.versions.delete(id)
            except CogniteNotFoundError:
                print(f"  [bold yellow]WARNING:[/] WorkflowVersion {id} does not exist, skipping delete.")
            else:
                successes += 1
        return successes

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[WorkflowVersion]:
        workflow_ids = [parent_id for parent_id in parent_ids if isinstance(parent_id, str)] if parent_ids else None
        return self.client.workflows.versions.list(limit=-1, workflow_version_ids=workflow_ids)  # type: ignore[arg-type]

    @classmethod
    def as_str(cls, id: WorkflowVersionId) -> str:
        if id.version is None:
            version = ""
        elif id.version.startswith("v"):
            version = f"_{id.version}"
        else:
            version = f"_v{id.version}"

        return sanitize_filename(f"{id.workflow_external_id}{version}")

    @classmethod
    def topological_sort(cls, items: Sequence[WorkflowVersionUpsert]) -> list[WorkflowVersionUpsert]:
        workflow_by_id: dict[WorkflowVersionId, WorkflowVersionUpsert] = {item.as_id(): item for item in items}
        dependencies: dict[WorkflowVersionId, set[WorkflowVersionId]] = {}
        for item_id, item in workflow_by_id.items():
            dependencies[item_id] = set()
            for task in item.workflow_definition.tasks:
                if isinstance(task.parameters, SubworkflowReferenceParameters):
                    dependencies[item_id].add(
                        WorkflowVersionId(task.parameters.workflow_external_id, task.parameters.version)
                    )

        try:
            return [
                workflow_by_id[item_id]
                for item_id in TopologicalSorter(dependencies).static_order()
                if item_id in workflow_by_id
            ]
        except CycleError as e:
            raise ToolkitCycleError(
                f"Cannot deploy workflows. Cycle detected {e.args} in the 'subworkflow' dependencies of the workflows.",
                *e.args[1:],
            ) from None


@final
class WorkflowTriggerCRUD(
    ResourceCRUD[str, WorkflowTriggerUpsert, WorkflowTrigger, WorkflowTriggerUpsertList, WorkflowTriggerList]
):
    folder_name = "workflows"
    filename_pattern = r"^.*WorkflowTrigger$"
    resource_cls = WorkflowTrigger
    resource_write_cls = WorkflowTriggerUpsert
    list_cls = WorkflowTriggerList
    list_write_cls = WorkflowTriggerUpsertList
    kind = "WorkflowTrigger"
    dependencies = frozenset({WorkflowCRUD, WorkflowVersionCRUD, GroupResourceScopedCRUD, GroupAllScopedCRUD})
    parent_resource = frozenset({WorkflowCRUD})
    yaml_cls = WorkflowTriggerYAML

    _doc_url = "Workflow-triggers/operation/CreateOrUpdateTriggers"

    class _MetadataKey:
        secret_hash = "cognite-toolkit-auth-hash"

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None = None):
        super().__init__(client, build_dir, console)
        self._authentication_by_id: dict[str, ClientCredentials] = {}

    @property
    def display_name(self) -> str:
        return "workflow triggers"

    @classmethod
    def get_id(cls, item: WorkflowTriggerUpsert | WorkflowTrigger | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[WorkflowTriggerUpsert] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        capability = (
            [WorkflowOrchestrationAcl.Action.Read]
            if read_only
            else [WorkflowOrchestrationAcl.Action.Read, WorkflowOrchestrationAcl.Action.Write]
        )

        return WorkflowOrchestrationAcl(
            capability,
            WorkflowOrchestrationAcl.Scope.All(),
        )

    def create(self, items: WorkflowTriggerUpsertList) -> WorkflowTriggerList:
        return self._upsert(items)

    def _upsert(self, items: WorkflowTriggerUpsertList) -> WorkflowTriggerList:
        created = WorkflowTriggerList([])
        for item in items:
            created.append(self._upsert_item(item))
        return created

    def _upsert_item(self, item: WorkflowTriggerUpsert) -> WorkflowTrigger:
        credentials = self._authentication_by_id.get(item.external_id)
        try:
            return self.client.workflows.triggers.upsert(item, credentials)
        except CogniteAuthError as e:
            if hint := try_find_error(credentials):
                raise ResourceCreationError(f"Failed to create WorkflowTrigger {item.external_id}: {hint}") from e
            raise e

    def retrieve(self, ids: SequenceNotStr[str]) -> WorkflowTriggerList:
        all_triggers = self.client.workflows.triggers.list(limit=-1)
        lookup = set(ids)
        return WorkflowTriggerList([trigger for trigger in all_triggers if trigger.external_id in lookup])

    def update(self, items: WorkflowTriggerUpsertList) -> WorkflowTriggerList:
        return self._upsert(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        successes = 0
        for id in ids:
            try:
                self.client.workflows.triggers.delete(external_id=id)
            except CogniteNotFoundError:
                LowSeverityWarning(f"WorkflowTrigger {id} does not exist, skipping delete.").print_warning()
            else:
                successes += 1
        return successes

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[WorkflowTrigger]:
        triggers = self.client.workflows.triggers.list(limit=-1)
        if parent_ids is not None:
            # Parent = Workflow
            workflow_ids = {parent_id for parent_id in parent_ids if isinstance(parent_id, str)}
            return (trigger for trigger in triggers if trigger.workflow_external_id in workflow_ids)
        return triggers

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires."""
        if "workflowExternalId" in item:
            yield WorkflowCRUD, item["workflowExternalId"]

            if "workflowVersion" in item:
                yield WorkflowVersionCRUD, WorkflowVersionId(item["workflowExternalId"], item["workflowVersion"])

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        resources = super().load_resource_file(filepath, environment_variables)

        # We need to the auth hash calculation here, as the output of the load_resource_file
        # is used to compare with the CDF resource.
        for resource in resources:
            identifier = self.get_id(resource)
            credentials = read_auth(
                resource.get("authentication"), self.client.config, identifier, "workflow trigger", console=self.console
            )
            self._authentication_by_id[identifier] = credentials
            if "metadata" not in resource:
                resource["metadata"] = {}
            resource["metadata"][self._MetadataKey.secret_hash] = calculate_secure_hash(
                credentials.dump(camel_case=True), shorten=True
            )
        return resources

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> WorkflowTriggerUpsert:
        if isinstance(resource.get("data"), dict):
            resource["data"] = json.dumps(resource["data"])
        return WorkflowTriggerUpsert._load(resource)

    def dump_resource(self, resource: WorkflowTrigger, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if isinstance(dumped.get("data"), str) and isinstance(local.get("data"), dict):
            dumped["data"] = json.loads(dumped["data"])

        cdf_rule = dumped.get("triggerRule", {})
        local_rule = local.get("triggerRule", {})
        if cdf_rule.get("triggerType") == "schedule" and local_rule.get("triggerType") == "schedule":
            if cdf_rule.get("timezone") == "UTC" and "timezone" not in local_rule:
                # The server defaults to UTC if not set, so we remove it if the local does not have it set.
                del cdf_rule["timezone"]

        if "authentication" in local:
            # Changes in auth will be detected by the hash. We need to do this to ensure
            # that the pull command works.
            dumped["authentication"] = local["authentication"]
        return dumped

    def sensitive_strings(self, item: WorkflowTriggerUpsert) -> Iterable[str]:
        id_ = self.get_id(item)
        if id_ in self._authentication_by_id:
            yield self._authentication_by_id[id_].client_secret



from cognite_toolkit._cdf_tk.commands import CleanCommand
from cognite_toolkit._cdf_tk.data_classes._config_yaml import BuildEnvironment


class TestCleanCommandSelectModules:
    """Test the CleanCommand._select_modules method."""

    def test_select_specific_module(
        self,
        build_environment: BuildEnvironment,
    ) -> None:
        """Test that when a specific module is specified, it is returned."""
        cmd = CleanCommand(silent=True, skip_tracking=True)
        result = cmd._select_modules(build_environment, module_str="my_example_module")
        assert len(result) == 1
        assert result[0].dir.name == "my_example_module", f"Expected 'my_example_module', got {result[0].dir.name}"



from collections.abc import Iterable
from unittest.mock import MagicMock

import pytest
from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes import ClientCredentials, OidcCredentials

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.exceptions import ToolkitError, ToolkitRequiredValueError, ToolkitTypeError
from cognite_toolkit._cdf_tk.utils.cdf import (
    get_transformation_destination_columns,
    get_transformation_sources,
    read_auth,
    try_find_error,
)


class TestTryFindError:
    @pytest.mark.parametrize(
        "credentials, expected",
        [
            pytest.param(
                ClientCredentials("${ENVIRONMENT_VAR}", "1234"),
                "The environment variable is not set: ENVIRONMENT_VAR.",
                id="Missing environment variable",
            ),
            pytest.param(
                ClientCredentials("${ENV1}", "${ENV2}"),
                "The environment variables are not set: ENV1 and ENV2.",
                id="Missing environment variable",
            ),
            pytest.param(
                OidcCredentials(
                    client_id="my-client-id",
                    client_secret="123",
                    scopes=["https://cognite.com"],
                    token_uri="not-valid-uri",
                    cdf_project_name="my-project",
                ),
                "The tokenUri 'not-valid-uri' is not a valid URI.",
            ),
            pytest.param(None, None, id="empty"),
        ],
    )
    def test_try_find_error(self, credentials: OidcCredentials | ClientCredentials | None, expected: str | None):
        assert try_find_error(credentials) == expected


def get_transformation_source_test_cases() -> Iterable:
    yield pytest.param(
        """
select
  identifier as externalId,
  name as name
from
    my_db.my_table""",
        [RawTable(db_name="my_db", table_name="my_table")],
        ["externalId", "name"],
        id="Simple query without joins",
    )

    yield pytest.param(
        """
        select table.identifier as externalId,
               table.name       as name
        from `my db`.`my table` as table""",
        [RawTable(db_name="my db", table_name="my table")],
        ["externalId", "name"],
        id="Use of backticks in table name with AS alias",
    )

    yield pytest.param(
        """SELECT
    table1.externalId as externalId,
    table2.name as name
FROM
    `ingestion`.`dump` as table1
LEFT JOIN
    `ingestion`.`workitem` as table2
ON table1.shared_column = table2.shared_column""",
        [RawTable(db_name="ingestion", table_name="dump"), RawTable(db_name="ingestion", table_name="workitem")],
        ["externalId", "name"],
        id="Query with left join",
    )

    yield pytest.param(
        """with parentLookup as (
  select
    concat('WMT:', cast(d1.`WMT_TAG_NAME` as STRING)) as externalId,
    node_reference('springfield_instances',  concat('WMT:', cast(d2.`WMT_TAG_NAME` as STRING))) as parent
  from
      `ingestion`.`dump` as  d1
  join
    `ingestion`.`dump` as d2
  on
    d1.`WMT_TAG_ID_ANCESTOR` = d2.`WMT_TAG_ID`
  where
    isnotnull(d1.`WMT_TAG_NAME`) AND
    cast(d1.`WMT_CATEGORY_ID` as INT) = 1157 AND
    isnotnull(d2.`WMT_TAG_NAME`) AND
    cast(d2.`WMT_CATEGORY_ID` as INT) = 1157
)
select
	concat('WMT:', cast(d3.`WMT_TAG_NAME` as STRING)) as externalId,
    parentLookup.parent,
    cast(`WMT_TAG_NAME` as STRING) as name,
    cast(`WMT_TAG_DESC` as STRING) as description,
    cast(`WMT_TAG_ID` as STRING) as sourceId,
    cast(`WMT_TAG_CREATED_DATE` as TIMESTAMP) as sourceCreatedTime,
    cast(`WMT_TAG_UPDATED_DATE` as TIMESTAMP) as sourceUpdatedTime,
    cast(`WMT_TAG_UPDATED_BY` as STRING) as sourceUpdatedUser
from
  `ingestion`.`dump` as d3
left join
	parentLookup
on
 concat('WMT:', cast(d3.`WMT_TAG_NAME` as STRING)) = parentLookup.externalId
where
  isnotnull(d3.`WMT_TAG_NAME`) AND
/* Inspection of the WMT_TAG_DESC looks like asset are category 1157 while equipment is everything else */
  cast(d3.`WMT_CATEGORY_ID` as INT) = 1157""",
        [RawTable(db_name="ingestion", table_name="dump")],
        [
            "externalId",
            "parent",
            "name",
            "description",
            "sourceId",
            "sourceCreatedTime",
            "sourceUpdatedTime",
            "sourceUpdatedUser",
        ],
        id="Nested query with join",
    )
    yield pytest.param(
        """select
  a.FL_0FUNCT_LOC as externalId,
  a.FL_0COSTCENTER as description,
  to_metadata(
    a.FL_0ABCINDIC,
    a.FL_0BEGRU,
    a.FL_0COMP_CODE,
  ) as metadata
from
  DTN.SAP_Functional_Location as a
  inner join (
    select
      *
    from
      DTN.SAP_Functional_Location
    where
      FL_ZFLOC_PAR NOT LIKE "ZZ%"
  ) as v on a.FL_ZFLOC_PAR = v.FL_0FUNCT_LOC
where
  is_new('my_sap_2123,', a.lastUpdatedTime)
  AND size(split(a.FL_ZFLOC_PAR, "-")) > 1
  AND a.FL_ZFLOC_PAR NOT LIKE "ZZ%"
  AND a.FL_0FUNCT_LOC NOT LIKE "TG-TGP-18%"
""",
        [RawTable(db_name="DTN", table_name="SAP_Functional_Location")],
        ["externalId", "description", "metadata"],
        id="Query with inner join",
    )

    yield pytest.param(
        """with
  asset_metadata_to_add as
    (select
        Vulnerability
  from `UpdateDB`.`Vulnerability`),

  exisiting_metadata as
    (select metadata, externalId from _cdf.assets where externalId in (select Floc from asset_metadata_to_add)),

  combined_table as
    (select asset_metadata_to_add.*,
        exisiting_metadata.*
      from asset_metadata_to_add
      JOIN exisiting_metadata
      on asset_metadata_to_add.Floc = exisiting_metadata.externalId)

select
  map_concat(to_metadata(
        Vulnerability),
        metadata) as metadata,
  combined_table.externalId as externalId
  from
  combined_table""",
        [RawTable(db_name="UpdateDB", table_name="Vulnerability"), "assets"],
        ["metadata", "externalId"],
        id="Source inside inner With Query",
    )

    yield pytest.param(
        """select
id as id,
externalId as externalId,
name as name,
1234 as dataSetId
from _cdf.assets
where left(externalId,5)="GM-ST" and left(name,2)="ST"
    """,
        ["assets"],
        ["id", "externalId", "name", "dataSetId"],
        id="Source is _cdf.assets (not a RawTable)",
    )
    yield pytest.param(
        """with
  created_assets as (
    select
      collect_set(externalId) as assets
    from
      _cdf.assets
    where
      dataSetId = 5238979812378
  ),
  bad_vals as (
    select
      array(
        '000000000012345678',
      ) as bv
  ),
  filtered_equipment as (
    select
      EQ_0EQUIPMENT as externalId
    from
      DTN.`SAP_EQUIPMENT`
  )
select
  *
from
  filtered_equipment
where
EQ_0FUNCT_LOC in (
    select
      explode(assets)
    from
      created_assets)
AND EQ_0EQUIPMENT not in (
  select
    explode(bv)
  from
    bad_vals
)
AND EQ_ZEQUI_PAR not in (
  select
    explode(bv)
  from
    bad_vals
)
""",
        ["assets", RawTable(db_name="DTN", table_name="SAP_EQUIPMENT")],
        ["*"],
        id="Source is a RawTable and _cdf.assets in a With Query",
    )

    yield pytest.param(
        """SELECT
  CASE
    WHEN REPLACE(LMD_ID, LMD.Component_ID, '') != LMD_ID THEN CONCAT(REPLACE(CP.parentExternalId, LMD.Component_ID, ''), LMD_ID)
    ELSE CONCAT(REPLACE(TRIM(CT.INT_CTRL_NO1), LMD.Circuit_ID, ''), LMD_ID)
  END AS externalId
  ,CASE
    WHEN LMD.Component_ID IS NULL THEN TRIM(CT.INT_CTRL_NO1)
    ELSE CP.parentExternalId
  END AS parentExternalId
  ,CASE
    WHEN LMD.Component_ID IS NULL THEN REPLACE(LMD_ID, CONCAT(LMD.Circuit_ID, '-'), '')
    ELSE REPLACE(LMD_ID, CONCAT(LMD.Component_ID, '-'), '')
  END AS name
  ,LMD.description AS description
  ,dataset_id('PSSD') AS dataSetId
  ,to_metadata_except(array('description', 'FACILITY'), LMD.*) AS metadata
  ,array('LMD') as labels
FROM (
  SELECT
    TRIM(EQUIP_ID) AS Equipment_ID
  FROM PSSD.PSSD_LMD
  WHERE is_new('pssd_lmd', lastUpdatedTime)) LMD
LEFT JOIN (
  SELECT
    CASE
      WHEN TRIM(`Internal Control No 1`) LIKE "%_E" THEN CONCAT(REPLACE(TRIM(ASSET_NO), TRIM(`EQUIPMENT ID`), ''), REPLACE(TRIM(`component_id`), ' ', ''))
      ELSE CONCAT(REPLACE(CL.INT_CTRL_NO1, TRIM(CIRCUIT_ID), ''), REPLACE(TRIM(`component_id`), ' ', ''))
    END AS parentExternalId,
    REPLACE(TRIM(`component_id`), ' ', '') AS  component_id,
    TRIM(`FACILITY ID`) AS FACILITY,
    TRIM(CPT.`EQUIPMENT ID`) AS EQUIP_ID
  FROM PSSD.PSSD_component CPT
  LEFT JOIN PSSD.PSSD_circuit CL
    ON TRIM(`Internal Control No 1`) = CL.CIRCUIT_ID
    AND TRIM(`FACILITY ID`) = FACILITY
    AND TRIM(CPT.`EQUIPMENT ID`) = TRIM(CL.EQUIP_ID)
  LEFT JOIN (
    SELECT externalId AS cdf_asset
    FROM _cdf.assets) CDF
      ON CASE
        WHEN TRIM(`Internal Control No 1`) LIKE "%_E" THEN TRIM(ASSET_NO)
        ELSE CL.INT_CTRL_NO1
      END = CDF.cdf_asset
  WHERE (TRIM(`Internal Control No 1`) LIKE "%_E"
    OR CIRCUIT_ID IS NOT NULL)
    AND cdf_asset IS NOT NULL
  ) CP
  ON TRIM(LMD.Component_ID) = TRIM(CP.component_id)
    AND TRIM(LMD.FACILITY) = TRIM(CP.FACILITY)
    AND LMD.Equipment_ID = CP.EQUIP_ID
    AND LMD.Circuit_ID = CP.CIRCUIT_ID
LEFT JOIN PSSD.PSSD_circuit CT
  ON TRIM(LMD.Circuit_ID) = TRIM(CT.CIRCUIT_ID)
  AND TRIM(LMD.FACILITY) = TRIM(CT.FACILITY)
  AND TRIM(LMD.Equipment_ID) = TRIM(CT.EQUIP_ID)
LEFT JOIN (
  SELECT externalId AS cdf_asset
  FROM _cdf.assets) CDF2
  ON CASE
    WHEN LMD.Component_ID IS NULL THEN TRIM(CT.INT_CTRL_NO1)
    ELSE CP.parentExternalId
  END = CDF2.cdf_asset
WHERE CP.component_id IS NOT NULL
  OR (CT.CIRCUIT_ID IS NOT NULL
    AND LMD.Component_ID IS NULL)
    AND cdf_asset IS NOT NULL
""",
        [RawTable("PSSD", "PSSD_LMD"), RawTable("PSSD", "PSSD_component"), RawTable("PSSD", "PSSD_circuit"), "assets"],
        ["externalId", "parentExternalId", "name", "description", "dataSetId", "metadata", "labels"],
        id="Complex query with multiple joins and CDF assets",
    )

    yield pytest.param(
        """SELECT
  id,
  map_concat(
    metadata,
    to_metadata(
        Station_Name,
        WellName,
        GIS_latitude,
        GIS_longitude,
        SPSOnLine,
        idwell
    )
  ) as metadata
FROM
  (
    select
      a.id,
      ws.Station_Name,
      sdo.WellName,
      gis.WI_LATITUDE as GIS_latitude,
      gis.WI_LONGITUDE as GIS_longitude,
      pw.SPSOnLine,
      wh.idwell
    FROM
      _cdf.assets as a
      left join SDO.`well_header` as sdo on a.metadata['PRANumber6Digits'] = sdo.PraNumber
      left join SDO.`pw_DepletionPlan` AS pw on a.metadata['PRANumber6Digits'] = pw.PRANumber6Digits
      left join SDO.`well_header2` as wh on substring(sdo.ApiNumber, 0, 10) = substring(wellida, 0, 10)
      left join `GIS`.`gis` as gis on substring(sdo.ApiNumber, 0, 10) = substring(WI_APINO, 0, 10)
      left join SDO.`weather_station` as ws on sdo.ApiNumber = ws.ApiNumber
    where
      array_contains(labels, "my_location_sap")
  )
    """,
        [
            "assets",
            RawTable(db_name="SDO", table_name="well_header"),
            RawTable(db_name="SDO", table_name="pw_DepletionPlan"),
            RawTable(db_name="SDO", table_name="well_header2"),
            RawTable(db_name="GIS", table_name="gis"),
            RawTable(db_name="SDO", table_name="weather_station"),
        ],
        ["id", "metadata"],
        id="Complex query with multiple joins and CDF assets with metadata",
    )

    yield pytest.param(
        """SELECT
  cast(concat("GM", `Tag name`) as STRING) as externalId,
  cast(`Parent tag` as STRING) as parentExternalId,
  cast(concat("TB-",`Tag name`) as STRING) as name,
  cast(concat(`Tag class name`," ", `Tag description`) as STRING) as description,
  1238456789 as dataSetId,
  'Engineering Documents' as source
FROM `GM_DATA`.`verified_new_only`

WHERE concat("GM", `Tag name`) IN (

    WITH existing_ext_ids AS (
        SELECT DISTINCT externalId AS value
        FROM _cdf.assets
    ),
    new_ext_ids AS (
        SELECT DISTINCT concat("GM", `Tag name`) AS value
        FROM `GM_DATA`.`verified_new_only`
    )
    SELECT COALESCE(existing_ext_ids.value, new_ext_ids.value) AS differing_value
    FROM existing_ext_ids
    RIGHT JOIN new_ext_ids
    ON existing_ext_ids.value = new_ext_ids.value
    WHERE existing_ext_ids.value IS NULL OR new_ext_ids.value IS NULL
    )
;""",
        [RawTable(db_name="GM_DATA", table_name="verified_new_only"), "assets"],
        ["externalId", "parentExternalId", "name", "description", "dataSetId", "source"],
        id="Query listed in WHERE clause",
    )

    yield pytest.param(
        """SELECT asset.id as id,
        array("SAP DELETED") as labels
        -- if(asset.labels is null, array("SAP DELETED"), array_union(asset.labels, array("SAP DELETED"))) as labels
FROM
(SELECT
CONCAT(REPLACE(EQ_0FUNCT_LOC, "-IN", ""), "-", key) AS ext_id
FROM DVT.`SAP_Equipment`
WHERE RLIKE(EQ_0MD,"FL")
UNION
SELECT
FL_0FUNCT_LOC AS ext_id
FROM DVT.SAP_Functional_Location
WHERE FL_PAR LIKE "ZZ%"
OR FL_SYS_ST != "CE"
UNION
SELECT
  FIRST(externalId) as ext_id
FROM (
  SELECT *,
   element_at(split(externalId, "-"), -1) as equip_id,
    COUNT(*) OVER (PARTITION BY element_at(split(externalId, "-"), -1)) AS count
  FROM `_cdf`.`assets`
  where dataSetId = dataset_id("sap_equipment")
  or dataSetId = dataset_id("midstream_equipment")
  order by createdTime asc
) tmp
WHERE count > 1
GROUP BY equip_id
)

INNER JOIN _cdf.assets asset on ext_id = asset.externalId
""",
        [
            RawTable(db_name="DVT", table_name="SAP_Equipment"),
            RawTable(db_name="DVT", table_name="SAP_Functional_Location"),
            "assets",
        ],
        ["id", "labels"],
        id="Query with UNION and multiple sources including CDF assets and a comment in the Select clause",
    )

    yield pytest.param(
        """
        select identifier as externalId,
            /* This is a comment */
               name       as name
        from my_db.my_table""",
        [RawTable(db_name="my_db", table_name="my_table")],
        ["externalId", "name"],
        id="Simple query with comment",
    )

    yield pytest.param(
        """with assetid (
select
  first(wa.sourceExternalId) as workorderid,
  collect_list(ai.id)        as assetIds
from
  `workorder_oid_workmate`.`workorder2assets` wa,
  _cdf.assets                                 ai
where
  ai.externalId = wa.targetExternalId
GROUP BY sourceExternalId)
select
  wo.externalId,
  wo.description,
  dataset_id('ds_transformations_oid')  as dataSetId,
  cast(from_unixtime(double(wo.`startTime`)/1000)
                         as TIMESTAMP)  as startTime,
  cast(from_unixtime(double(wo.`endTime`)/1000)
                         as TIMESTAMP)  as endTime,
  ai.assetIds                           as assetIds,
  "workorder"                           as type,
  "OID - workmate"                      as source,
  to_metadata_except(
    array("key",
          "startTime",
          "endTime",
          "externalId",
          "wo.description",
          "assetIds"), *)              as metadata
from
  `workorder_oid_workmate`.`workorders` wo,
  assetid ai
where
  ai.workorderid = wo.workOrderNumber
  and wo.`endTime` > wo.`startTime`""",
        [
            RawTable(db_name="workorder_oid_workmate", table_name="workorder2assets"),
            "assets",
            RawTable(db_name="workorder_oid_workmate", table_name="workorders"),
        ],
        [
            "externalId",
            "description",
            "dataSetId",
            "startTime",
            "endTime",
            "assetIds",
            "type",
            "source",
            "metadata",
        ],
        id="Query with listed from.",
    )


class TestGetTransformationSource:
    @pytest.mark.parametrize(
        "query, expected_sources, expected_destination_columns", list(get_transformation_source_test_cases())
    )
    def test_get_transformation_sources(
        self, query: str, expected_sources: list[RawTable | str], expected_destination_columns: list[str]
    ) -> None:
        """Test that the transformation source is correctly extracted from the query."""
        actual = get_transformation_sources(query)
        assert actual == expected_sources


class TestGetTransformationDestinationColumns:
    @pytest.mark.parametrize(
        "query, expected_sources, expected_destination_columns",
        list(get_transformation_source_test_cases()),
    )
    def test_get_transformation_sources_and_destination_columns(
        self, query: str, expected_sources: list[RawTable | str], expected_destination_columns: list[str]
    ) -> None:
        """Test that the transformation source and destination columns are correctly extracted from the query."""
        actual = get_transformation_destination_columns(query)
        assert actual == expected_destination_columns


class TestReadAuth:
    @pytest.mark.parametrize(
        "auth, expected",
        [
            pytest.param(
                {
                    "clientId": "my_id",
                    "clientSecret": "my_secret",
                },
                ClientCredentials("my_id", "my_secret"),
                id="Client credentials",
            ),
            pytest.param(
                {
                    "clientId": "my_id",
                    "clientSecret": "my_secret",
                    "tokenUri": "https://my-token-uri",
                    "cdfProjectName": "my-project",
                },
                OidcCredentials("my_id", "my_secret", "https://my-token-uri", "my-project"),
                id="OIDC credentials only required",
            ),
            pytest.param(
                {
                    "clientId": "my_id",
                    "clientSecret": "my_secret",
                    "tokenUri": "https://my-token-uri",
                    "cdfProjectName": "my-project",
                    "scopes": "USER_IMPERSONATION,https://cognite.com",
                    "audience": "https://cognite.com",
                },
                OidcCredentials(
                    "my_id",
                    "my_secret",
                    "https://my-token-uri",
                    "my-project",
                    ["USER_IMPERSONATION", "https://cognite.com"],
                    "https://cognite.com",
                ),
                id="OIDC credentials all fields",
            ),
        ],
    )
    def test_read_valid_auth(self, auth: object, expected: ClientCredentials | OidcCredentials) -> None:
        config = MagicMock(spec=ToolkitClientConfig)
        result = read_auth(auth, config, "only-used-in-errors", "only-used-in-errors", allow_oidc=True)
        assert isinstance(result, ClientCredentials | OidcCredentials)
        assert result.dump() == expected.dump()

    @pytest.mark.parametrize(
        "auth, expected_exception",
        [
            pytest.param(
                None,
                ToolkitRequiredValueError("Authentication is missing for compute resource 'my_compute_resource'."),
                id="Missing authentication",
            ),
            pytest.param(
                123,
                ToolkitTypeError("Authentication must be a dictionary for compute resource 'my_compute_resource'"),
            ),
            pytest.param(
                {"clientId": "my_id"},
                ToolkitRequiredValueError(
                    "Authentication must contain clientId and clientSecret for compute resource 'my_compute_resource'"
                ),
            ),
        ],
    )
    def test_read_invalid_auth(self, auth: object, expected_exception: ToolkitError) -> None:
        with pytest.raises(type(expected_exception)) as excinfo:
            config = MagicMock(spec=ToolkitClientConfig)
            config.is_strict_validation = True
            read_auth(auth, config, "my_compute_resource", "compute resource")

        assert str(excinfo.value) == str(expected_exception)

    def test_read_warning_auth(self) -> None:
        credentials = OAuthClientCredentials("url", "my_id", "my_secret", ["USER_IMPERSONATION"])
        config = MagicMock(spec=ToolkitClientConfig)
        config.is_strict_validation = False
        config.credentials = credentials
        warning: str = ""

        def catch_warning_message(*messages: object) -> None:
            nonlocal warning
            warning = "".join(map(str, messages))

        console = MagicMock()
        console.print = catch_warning_message

        result = read_auth(None, config, "my_compute_resource", "compute resource", console=console)
        assert (
            "Authentication is missing for compute resource 'my_compute_resource'. "
            "Falling back to the Toolkit credentials"
        ) in warning
        assert isinstance(result, ClientCredentials)
        assert result.dump() == {
            "clientId": "my_id",
            "clientSecret": "my_secret",
        }



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from typing import Literal, TypeAlias

from cognite_toolkit._cdf_tk.feature_flags import FeatureFlag, Flags

from ._base_cruds import DataCRUD, Loader, ResourceContainerCRUD, ResourceCRUD
from ._data_cruds import DatapointsCRUD, FileCRUD, RawFileCRUD
from ._resource_cruds import (
    AgentCRUD,
    AssetCRUD,
    CogniteFileCRUD,
    ContainerCRUD,
    DataModelCRUD,
    DatapointSubscriptionCRUD,
    DataSetsCRUD,
    EdgeCRUD,
    EventCRUD,
    ExtractionPipelineConfigCRUD,
    ExtractionPipelineCRUD,
    FileMetadataCRUD,
    FunctionCRUD,
    FunctionScheduleCRUD,
    GraphQLCRUD,
    GroupAllScopedCRUD,
    GroupCRUD,
    GroupResourceScopedCRUD,
    HostedExtractorDestinationCRUD,
    HostedExtractorJobCRUD,
    HostedExtractorMappingCRUD,
    HostedExtractorSourceCRUD,
    InFieldLocationConfigCRUD,
    InfieldV1CRUD,
    LabelCRUD,
    LocationFilterCRUD,
    NodeCRUD,
    RawDatabaseCRUD,
    RawTableCRUD,
    RelationshipCRUD,
    ResourceViewMappingCRUD,
    RobotCapabilityCRUD,
    RoboticFrameCRUD,
    RoboticLocationCRUD,
    RoboticMapCRUD,
    RoboticsDataPostProcessingCRUD,
    SearchConfigCRUD,
    SecurityCategoryCRUD,
    SequenceCRUD,
    SequenceRowCRUD,
    SpaceCRUD,
    StreamlitCRUD,
    ThreeDModelCRUD,
    TimeSeriesCRUD,
    TransformationCRUD,
    TransformationNotificationCRUD,
    TransformationScheduleCRUD,
    ViewCRUD,
    WorkflowCRUD,
    WorkflowTriggerCRUD,
    WorkflowVersionCRUD,
)
from ._worker import ResourceWorker

_EXCLUDED_CRUDS: set[type[ResourceCRUD]] = set()
if not FeatureFlag.is_enabled(Flags.GRAPHQL):
    _EXCLUDED_CRUDS.add(GraphQLCRUD)
if not FeatureFlag.is_enabled(Flags.AGENTS):
    _EXCLUDED_CRUDS.add(AgentCRUD)
if not FeatureFlag.is_enabled(Flags.INFIELD):
    _EXCLUDED_CRUDS.add(InfieldV1CRUD)
    _EXCLUDED_CRUDS.add(InFieldLocationConfigCRUD)
if not FeatureFlag.is_enabled(Flags.MIGRATE):
    _EXCLUDED_CRUDS.add(ResourceViewMappingCRUD)
if not FeatureFlag.is_enabled(Flags.SEARCH_CONFIG):
    _EXCLUDED_CRUDS.add(SearchConfigCRUD)


CRUDS_BY_FOLDER_NAME: dict[str, list[type[Loader]]] = {}
for _loader in itertools.chain(
    ResourceCRUD.__subclasses__(),
    ResourceContainerCRUD.__subclasses__(),
    DataCRUD.__subclasses__(),
    GroupCRUD.__subclasses__(),
):
    if _loader in [ResourceCRUD, ResourceContainerCRUD, DataCRUD, GroupCRUD] or _loader in _EXCLUDED_CRUDS:
        # Skipping base classes
        continue
    if _loader.folder_name not in CRUDS_BY_FOLDER_NAME:  # type: ignore[attr-defined]
        CRUDS_BY_FOLDER_NAME[_loader.folder_name] = []  # type: ignore[attr-defined]
    # MyPy bug: https://github.com/python/mypy/issues/4717
    CRUDS_BY_FOLDER_NAME[_loader.folder_name].append(_loader)  # type: ignore[arg-type, attr-defined]
del _loader  # cleanup module namespace

CRUD_LIST = list(itertools.chain.from_iterable(CRUDS_BY_FOLDER_NAME.values()))
RESOURCE_CRUD_LIST = [loader for loader in CRUD_LIST if issubclass(loader, ResourceCRUD)]
RESOURCE_CRUD_CONTAINER_LIST = [loader for loader in CRUD_LIST if issubclass(loader, ResourceContainerCRUD)]
RESOURCE_DATA_CRUD_LIST = [loader for loader in CRUD_LIST if issubclass(loader, DataCRUD)]
KINDS_BY_FOLDER_NAME: dict[str, set[str]] = {}
for crud in CRUD_LIST:
    if crud.folder_name not in KINDS_BY_FOLDER_NAME:
        KINDS_BY_FOLDER_NAME[crud.folder_name] = set()
    KINDS_BY_FOLDER_NAME[crud.folder_name].add(crud.kind)
del crud  # cleanup module namespace

ResourceTypes: TypeAlias = Literal[
    "3dmodels",
    "agents",
    "auth",
    "cdf_applications",
    "classic",
    "data_models",
    "data_sets",
    "hosted_extractors",
    "locations",
    "migration",
    "transformations",
    "files",
    "timeseries",
    "extraction_pipelines",
    "functions",
    "raw",
    "robotics",
    "streamlit",
    "workflows",
]


def get_crud(resource_dir: str, kind: str) -> type[Loader]:
    for loader in CRUDS_BY_FOLDER_NAME[resource_dir]:
        if loader.kind == kind:
            return loader
    raise ValueError(f"Loader not found for {resource_dir} and {kind}")


__all__ = [
    "CRUDS_BY_FOLDER_NAME",
    "CRUD_LIST",
    "KINDS_BY_FOLDER_NAME",
    "RESOURCE_CRUD_CONTAINER_LIST",
    "RESOURCE_CRUD_LIST",
    "RESOURCE_DATA_CRUD_LIST",
    "AssetCRUD",
    "CogniteFileCRUD",
    "ContainerCRUD",
    "DataCRUD",
    "DataModelCRUD",
    "DataSetsCRUD",
    "DatapointSubscriptionCRUD",
    "DatapointsCRUD",
    "EdgeCRUD",
    "EventCRUD",
    "ExtractionPipelineCRUD",
    "ExtractionPipelineConfigCRUD",
    "FileCRUD",
    "FileMetadataCRUD",
    "FunctionCRUD",
    "FunctionScheduleCRUD",
    "GroupAllScopedCRUD",
    "GroupCRUD",
    "GroupResourceScopedCRUD",
    "HostedExtractorDestinationCRUD",
    "HostedExtractorJobCRUD",
    "HostedExtractorMappingCRUD",
    "HostedExtractorSourceCRUD",
    "InFieldLocationConfigCRUD",
    "LabelCRUD",
    "LocationFilterCRUD",
    "NodeCRUD",
    "RawDatabaseCRUD",
    "RawFileCRUD",
    "RawTableCRUD",
    "RelationshipCRUD",
    "ResourceCRUD",
    "ResourceContainerCRUD",
    "ResourceTypes",
    "ResourceWorker",
    "RobotCapabilityCRUD",
    "RoboticFrameCRUD",
    "RoboticLocationCRUD",
    "RoboticMapCRUD",
    "RoboticsDataPostProcessingCRUD",
    "SearchConfigCRUD",
    "SecurityCategoryCRUD",
    "SequenceCRUD",
    "SequenceRowCRUD",
    "SpaceCRUD",
    "StreamlitCRUD",
    "ThreeDModelCRUD",
    "TimeSeriesCRUD",
    "TransformationCRUD",
    "TransformationNotificationCRUD",
    "TransformationScheduleCRUD",
    "ViewCRUD",
    "WorkflowCRUD",
    "WorkflowTriggerCRUD",
    "WorkflowVersionCRUD",
    "get_crud",
]



"""These tests are for the dev.py CLI used to create the bump and changelog entry"""

from pathlib import Path
from unittest.mock import MagicMock

import pytest

import dev


def get_release_process_test_cases():
    yield pytest.param(
        """# Description

Use the new `pydantic` class to replace the `get_paramter_spec` for giving error message. So far we only support `TimeSeries`.

## Changelog

- [X] Patch
- [ ] Minor
- [ ] Skip

## cdf

### Improved

- Improved warnings if a `TimeSeries` configuration is not following the expected spec.

## templates

No changes.""",
        "0.5.15",
        """## cdf

### Improved

- Improved warnings if a `TimeSeries` configuration is not following the expected spec.

## templates

No changes.""",
        "0.5.16",
        id="Patch bump",
    )

    yield pytest.param(
        """# Description

Release `0.5.0`. This is done by removing the alpha flags for the 4 features that are planned for v5.

## Changelog

- [ ] Patch
- [x] Minor
- [ ] Skip

## cdf

### Changed

- The `cdf dump datamodel` no longer includes data models that are global (in system space). You can get the original behavior by using the `--include-global` flag.
- Toolkit now stores a hash of the credentials of Workflow/Transformation/Function in the resources such that the resource is updated when the credentials change.
- For Workflow/Transformations/Function Toolkit no longer falls back to Toolkit credentials when validation-type != 'dev' in the `config.[env].yaml`.

### Added
- Toolkit now support for `cdf dump workflow/transformation/group/node`.

## templates

No changes.
---------

Co-authored-by: Member <member@users.noreply.github.com>
Co-authored-by: Member <member@cognite.com>""",
        "0.4.39",
        """## cdf

### Changed

- The `cdf dump datamodel` no longer includes data models that are global (in system space). You can get the original behavior by using the `--include-global` flag.
- Toolkit now stores a hash of the credentials of Workflow/Transformation/Function in the resources such that the resource is updated when the credentials change.
- For Workflow/Transformations/Function Toolkit no longer falls back to Toolkit credentials when validation-type != 'dev' in the `config.[env].yaml`.

### Added
- Toolkit now support for `cdf dump workflow/transformation/group/node`.

## templates

No changes.""",
        "0.5.0",
        id="Minor bump with co-authors",
    )

    yield pytest.param(
        """This PR contains the following updates:

| Package | Change | Age | Adoption | Passing | Confidence |
|---|---|---|---|---|---|
| [certifi](https://redirect.github.com/certifi/python-certifi) | `==2023.11.17` -> `==2024.7.4` | [![age](https://developer.mend.io/api/mc/badges/age/pypi/certifi/2024.7.4?slim=true)](https://docs.renovatebot.com/merge-confidence/) | [![adoption](https://developer.mend.io/api/mc/badges/adoption/pypi/certifi/2024.7.4?slim=true)](https://docs.renovatebot.com/merge-confidence/) | [![passing](https://developer.mend.io/api/mc/badges/compatibility/pypi/certifi/2023.11.17/2024.7.4?slim=true)](https://docs.renovatebot.com/merge-confidence/) | [![confidence](https://developer.mend.io/api/mc/badges/confidence/pypi/certifi/2023.11.17/2024.7.4?slim=true)](https://docs.renovatebot.com/merge-confidence/) |

### GitHub Vulnerability Alerts

#### [CVE-2024-39689](https://redirect.github.com/certifi/python-certifi/security/advisories/GHSA-248v-346w-9cwc)

Certifi 2024.07.04 removes root certificates from "GLOBALTRUST" from the root store. These are in the process of being removed from Mozilla's trust store.

GLOBALTRUST's root certificates are being removed pursuant to an investigation which identified "long-running and unresolved compliance issues". Conclusions of Mozilla's investigation can be found [here]( https://groups.google.com/a/mozilla.org/g/dev-security-policy/c/XpknYMPO8dI).

---

### Release Notes

<details>
<summary>certifi/python-certifi (certifi)</summary>

### [`v2024.7.4`](https://redirect.github.com/certifi/python-certifi/compare/2024.06.02...2024.07.04)

[Compare Source](https://redirect.github.com/certifi/python-certifi/compare/2024.06.02...2024.07.04)

### [`v2024.6.2`](https://redirect.github.com/certifi/python-certifi/compare/2024.02.02...2024.06.02)

[Compare Source](https://redirect.github.com/certifi/python-certifi/compare/2024.02.02...2024.06.02)

### [`v2024.2.2`](https://redirect.github.com/certifi/python-certifi/compare/2023.11.17...2024.02.02)

[Compare Source](https://redirect.github.com/certifi/python-certifi/compare/2023.11.17...2024.02.02)

</details>

---

### Configuration

📅 **Schedule**: Branch creation - "" (UTC), Automerge - At any time (no schedule defined).

🚦 **Automerge**: Disabled by config. Please merge this manually once you are satisfied.

♻ **Rebasing**: Whenever PR becomes conflicted, or you tick the rebase/retry checkbox.

🔕 **Ignore**: Close this PR and you won't be reminded about this update again.

---

 - [ ] <!-- rebase-check -->If you want to rebase/retry this PR, check this box

---

This PR was generated by [Mend Renovate](https://mend.io/renovate/). View the [repository job log](https://developer.mend.io/github/cognitedata/toolkit).
<!--renovate-debug:eyJjcmVhdGVkSW5WZXIiOiI0MC4xNi4wIiwidXBkYXRlZEluVmVyIjoiNDAuMTYuMCIsInRhcmdldEJyYW5jaCI6Im1haW4iLCJsYWJlbHMiOlsiZGVwZW5kZW5jaWVzIiwicmVub3ZhdGUiXX0=-->""",
        "0.5.16",
        "",
        "0.5.16",
        id="Renovate PR",
    )

    yield pytest.param(
        """fix configuration value for workflows, fix bug with datapoints extractor (#1580)
Note to risk reviewer: the code in this PR is not part of the Toolkit
CLI per se, but rather a sample used by the Academy team when they run
Bootcamps. Hence the requirements for testing and typing etc are a bit
more relaxed.

# Description

Fixed Bootcamp workflows using incorrect credentials, fixed bug
introduced in last change for datapoints extractor.

## Changelog

- [x] Patch
- [ ] Minor
- [ ] Skip

## cdf

No changes

## templates

### Fixed

- Bootcamp config.test.yaml had incorrect credentials for workflow
variables
- fixed bug that would miss list of datapoints when list was < 50""",
        "0.5.16",
        """## cdf

No changes

## templates

### Fixed

- Bootcamp config.test.yaml had incorrect credentials for workflow
variables
- fixed bug that would miss list of datapoints when list was < 50""",
        "0.5.17",
        id="Patch bump with only templates changes",
    )


class TestReleaseProcess:
    @pytest.mark.parametrize(
        "last_git_message, last_version, expected_changelog, expected_version", list(get_release_process_test_cases())
    )
    def test_bump_and_create_changelog_entry(
        self, last_git_message: str, last_version: str, expected_changelog: str, expected_version: str, monkeypatch
    ) -> None:
        actual_changelog_entry: str | None = None
        actual_version: str | None = None
        last_git_message_file = MagicMock(spec=Path)
        last_git_message_file.read_text.return_value = last_git_message
        last_version_file = MagicMock(spec=Path)
        last_version_file.read_text.return_value = last_version

        def mock_write_changelog(content, encoding=None):
            nonlocal actual_changelog_entry
            actual_changelog_entry = content

        changelog_file = MagicMock(spec=Path)
        changelog_file.write_text = mock_write_changelog

        version_file = MagicMock(spec=Path)
        version_file.read_text.return_value = dev.VERSION_PLACEHOLDER

        def mock_write_version(content, **_):
            nonlocal actual_version
            actual_version = content

        version_file.write_text = mock_write_version

        monkeypatch.setattr(dev, "LAST_GIT_MESSAGE_FILE", last_git_message_file)
        monkeypatch.setattr(dev, "LAST_VERSION", last_version_file)
        monkeypatch.setattr(dev, "CHANGELOG_ENTRY_FILE", changelog_file)
        monkeypatch.setattr(dev, "VERSION_FILES", [version_file])
        monkeypatch.setattr(dev, "DOCKER_IMAGE_FILES", [])

        dev.bump()
        dev.create_changelog_entry()

        is_skip = last_version == expected_version
        if is_skip:
            assert actual_changelog_entry is None
        else:
            assert actual_changelog_entry is not None, "Changelog entry was not created"
            removed_trailing_space = "\n".join([line.rstrip() for line in actual_changelog_entry.splitlines()])
            assert removed_trailing_space == expected_changelog

            assert actual_version is not None, "Version was not updated"
            assert actual_version == expected_version



from typing import Literal

from pydantic import Field

from .agent_tools import AgentTool
from .base import ToolkitResource

Model = Literal[
    "azure/o3",
    "azure/o4-mini",
    "azure/gpt-4o",
    "azure/gpt-4o-mini",
    "azure/gpt-4.1",
    "azure/gpt-4.1-nano",
    "azure/gpt-4.1-mini",
    "azure/gpt-5",
    "azure/gpt-5-mini",
    "azure/gpt-5-nano",
    "gcp/gemini-2.5-pro",
    "gcp/gemini-2.5-flash",
    "aws/claude-4-sonnet",
    "aws/claude-4-opus",
    "aws/claude-4.1-opus",
    "aws/claude-3.5-sonnet",
]


class AgentYAML(ToolkitResource):
    """Atlas AI Agent"""

    external_id: str = Field(
        description="An external ID that uniquely identifies the agent.",
        min_length=1,
        max_length=128,
        pattern=r"^[^\x00]{1,128}$",
    )
    name: str = Field(
        description="A descriptive name intended for use in user interfaces.",
        min_length=1,
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="A human-readable description of what the agent does, used for documentation only. "
        "This description is not used by the language model.",
        max_length=1024,
    )
    instructions: str | None = Field(
        default=None,
        description="The instructions for the agent prompt the language model to understand "
        "the agent's goals and how to achieve them.",
        max_length=32000,
    )
    model: Model = Field(
        "azure/gpt-4o-mini", description="The name of the model to use. Defaults to your CDF project's default model."
    )
    tools: list[AgentTool] | None = Field(None, description="A list of tools available to the agent.", max_length=20)
    runtime_version: str | None = Field(None, description="The runtime version")



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.extraction_pipeline_config import ExtractionPipelineConfigYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_extraction_pipeline_config_test_cases() -> Iterable:
    # Missing required field: config
    yield pytest.param(
        {"externalId": "myConfig"}, {"Missing required field: 'config'"}, id="Missing required field: config"
    )
    # Missing required field: external_id
    yield pytest.param(
        {"config": {"key": "value"}}, {"Missing required field: 'externalId'"}, id="Missing required field: externalId"
    )
    # Empty external_id
    yield pytest.param(
        {"externalId": "", "config": {"key": "value"}},
        {"In field externalId string should have at least 1 character"},
        id="Empty externalId",
    )
    # config is not a string or dict
    yield pytest.param(
        {"externalId": "myConfig", "config": 123},
        {
            "In config.dict input should be a valid dictionary. Got 123 of type int.",
            "In config.str input should be a valid string. Got 123 of type int. Hint: "
            "Use double quotes to force string.",
        },
        id="Config not string or dict",
    )
    # Unknown field present
    yield pytest.param(
        {"externalId": "myConfig", "config": {"key": "value"}, "unknownField": 1},
        {"Unused field: 'unknownField'"},
        id="Unknown field present",
    )


class TestExtractionPipelineConfigYAML:
    @pytest.mark.parametrize("data", list(find_resources("Config", resource_dir="extraction_pipelines")))
    def test_load_valid_asset(self, data: dict[str, object]) -> None:
        loaded = ExtractionPipelineConfigYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_extraction_pipeline_config_test_cases()))
    def test_invalid_asset_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, ExtractionPipelineConfigYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors
